from dataclasses import dataclass, field

from hotsos.core.host_helpers import (
    InstallInfoBase,
    SnapPackageHelper,
    SystemdHelper,
)
from hotsos.core.plugintools import PluginPartBase

SNAP_LIST = ['microceph', 'microovn', 'microcloud']
CORE_SNAPS = [rf"(?:snap\.)?{p}" for p in SNAP_LIST]
DEP_SNAPS = ['lxd']
SERVICE_EXPRS = [rf"{s}\S*" for s in CORE_SNAPS]


@dataclass
class MicroCloudInstallInfo(InstallInfoBase):
    """ MicroCloud installation information. """
    snaps: SnapPackageHelper = field(default_factory=lambda:
                                     SnapPackageHelper(core_snaps=CORE_SNAPS,
                                                       other_snaps=DEP_SNAPS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(service_exprs=SERVICE_EXPRS))


class MicroCloudChecks(PluginPartBase):
    """ MicroCloud Checks. """
    plugin_name = 'microcloud'
    plugin_root_index = 15

    def __init__(self, *args, **kwargs):
        super().__init__()
        MicroCloudInstallInfo().mixin(self)

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        if MicroCloudInstallInfo().snaps.core:
            return True

        return False



# flake8: noqa
# pylint: disable=C0301

# Including this as a dep of (at least) Neutron
# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' ovsdbapp/exceptions.py
OVSDBAPP_EXCEPTIONS = [
    "OvsdbAppException",
    "TimeoutException",
    "OvsdbConnectionUnavailable",
]



import abc
import os
from collections import OrderedDict, UserList

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import SYSCtlFactory, CLIHelperFile
from hotsos.core.log import log
from hotsos.core.search import FileSearcher, SearchDef, ResultFieldInfo


class ProcNetBase(abc.ABC):
    """
    Provides a common way to extract fields from /proc/net/snmp and netstat.

    Expected file format is:

    hdr1: label1 label2 ...
    hdr1: value1 value2 ...
    hdr2: label3 label4 ...
    hdr2: value3 value4 ...
    """
    def __init__(self):
        self._data = {}

    def _pcent_of(self, field_count, total):
        """
        @param field_count: field we want to determine as percentage if total
        @param total: total value
        """
        count = getattr(self, field_count)
        if not count:
            return 0

        if not total:
            return 0

        return round((count * 100.0) / total, 2)

    def process_file(self, fname):
        if not os.path.exists(fname):
            log.debug("file not found '%s' - skipping load", fname)
            return

        log.debug("start processing %s", fname)
        with open(fname, encoding='utf-8') as fd:
            labels = None
            values = None
            for line in fd:
                if labels is None:
                    labels = line
                    continue

                values = line
                if not all([labels, values]):
                    break

                labels = labels.split()
                values = values.split()
                if labels[0] != values[0]:
                    log.warning('badness: header mismatch in file %s', fname)
                    break

                hdr = labels[0].rstrip(':')
                if hdr in self._data:
                    log.warning('badness: duplicate header in file %s', fname)
                    break

                self._data[hdr] = {}
                for i in list(range(1, len(labels))):
                    self._data[hdr][labels[i]] = int(values[i])

                labels = None
                values = None

        log.debug("finished processing %s", fname)

    def _get_field(self, header, field):
        """Return counter value for field in section header, or
        zero if field does not exist (because older kernel versions
        do not have some fields).
        """
        return self._data.get(header, {}).get(field, 0)

    @property
    @abc.abstractmethod
    def _header(self):
        """
        """

    @property
    @abc.abstractmethod
    def _fields(self):
        """
        """

    def __getattr__(self, fld):
        if fld in self._fields:
            return self._get_field(self._header, fld)

        raise AttributeError(f"{fld} not found in {self.__class__.__name__}")


class SNMPBase(ProcNetBase):
    """ Base class for /proc/net/snmp interface implementations. """
    def __init__(self):
        super().__init__()
        self.process_file(os.path.join(HotSOSConfig.data_root,
                                       'proc/net/snmp'))


class SNMPTcp(SNMPBase):
    """ /proc/net/snmp interface implementation to extract TCP information. """
    def _percent_in_segs(self, field):
        """
        The value of a field can be provided as a percentage of the total rx
        segments.
        """
        return self._pcent_of(field, self.InSegs)

    def _percent_out_segs(self, field):
        """
        The value of a field can be provided as a percentage of the total tx
        segments.
        """
        return self._pcent_of(field, self.OutSegs)

    @property
    def _header(self):
        return 'Tcp'

    @property
    def _fields(self):
        return ['InCsumErrors',
                # number of TCP segments received
                'InSegs',
                # number of TCP segments retransmitted
                'RetransSegs',
                # number of TCP segments sent
                'OutSegs']

    def __getattr__(self, fld):
        """
        Fields can be appended with PcentInSegs/PcentOutSegs to get a
        percentage of total rx/tx segments.
        """
        if fld.endswith('PcentInSegs'):
            fld = fld.partition('PcentInSegs')[0]
            return self._percent_in_segs(fld)
        if fld.endswith('PcentOutSegs'):
            fld = fld.partition('PcentOutSegs')[0]
            return self._percent_out_segs(fld)

        return super().__getattr__(fld)


class SNMPUdp(SNMPBase):
    """ /proc/net/snmp interface implementation to extract UDP information. """
    @property
    def _header(self):
        return 'Udp'

    def _percent_in_datagrams(self, field):
        """
        The value of a field can be provided as a percentage of the total rx
        datagrams.
        """
        return self._pcent_of(field, self.InDatagrams)

    def _percent_out_datagrams(self, field):
        """
        The value of a field can be provided as a percentage of the total tx
        datagrams.
        """
        return self._pcent_of(field, self.OutDatagrams)

    @property
    def _fields(self):
        return ['InDatagrams',
                'NoPorts',
                'InErrors',
                'OutDatagrams',
                'RcvbufErrors',
                'SndbufErrors',
                'InCsumErrors']

    def __getattr__(self, fld):
        """
        Fields can be appended with PcentInDatagrams/PcentOutDatagrams to get a
        percentage of total rx/tx datagrams.
        """
        if fld.endswith('PcentInDatagrams'):
            fld = fld.partition('PcentInDatagrams')[0]
            return self._percent_in_datagrams(fld)
        if fld.endswith('PcentOutDatagrams'):
            fld = fld.partition('PcentOutDatagrams')[0]
            return self._percent_out_datagrams(fld)

        return super().__getattr__(fld)


class NetStatBase(ProcNetBase):
    """ Base class for /proc/net/netstat implementations.  """
    def __init__(self):
        super().__init__()
        self.process_file(os.path.join(HotSOSConfig.data_root,
                                       'proc/net/netstat'))
        self.net_snmp_tcp = SNMPTcp()


class NetStatTCP(NetStatBase):
    """
    /proc/net/netstat interface implementation to extract TCP information.
    """
    @property
    def _header(self):
        return 'TcpExt'

    def _percent_in_segs(self, field):
        """
        The value of a field can be provided as a percentage of the total rx
        segments.
        """
        return self._pcent_of(field, self.net_snmp_tcp.InSegs)

    def _percent_out_segs(self, field):
        """
        The value of a field can be provided as a percentage of the total tx
        segments.
        """
        return self._pcent_of(field, self.net_snmp_tcp.OutSegs)

    @property
    def _fields(self):
        return [
            # Times tried to reduce socket memory usage
            'PruneCalled',
            # of skbs collapsed (data combined)
            'TCPRcvCollapsed',
            # Times instructed TCP to drop due to mem press
            'RcvPruned',
            # Times threw away data in ofo queue
            'OfoPruned',
            # Adding skb to TCP backlog queue or C/R obscure case (TCP_REPAIR)
            'TCPBacklogDrop',
            # PFMEMALLOC skb to !MEMALLOC socket or C/R obscure case
            # (TCP_REPAIR).
            'PFMemallocDrop',
            # IP TTL < min TTL (default min TTL == 0) or C/R obscure case
            # (TCP_REPAIR).
            'TCPMinTTLDrop',
            # bare ACK in TCP_DEFER_ACCEPT mode (harmless) or C/R obscure case
            # (TCP_REPAIR)
            'TCPDeferAcceptDrop',
            # Drop after fail rp_filter check
            'IPReversePathFilter',
            # TCP accept queue overflow
            'ListenOverflows',
            # Catch-all for TCP incoming conn request drops or C/R obscure case
            # (TCP_REPAIR)
            'ListenDrops',
            # Drop if OFO and no rmem available for socket
            'TCPOFODrop',
            # Drop due to TCP recv window closed
            'TCPZeroWindowDrop',
            # No rmem when recv segment in ESTABLISHED state or C/R obscure
            # case (TCP_REPAIR).
            'TCPRcvQDrop',
            # req sock queue full (syn flood, syncookies off)
            'TCPReqQFullDrop',
            # req sock queue full (syn flood, syncookies on)
            'TCPReqQFullDoCookies',
            # skb retrans before original left host
            'TCPSpuriousRtxHostQueues']

    def __getattr__(self, fld):
        """
        Fields can be appended with PcentInSegs/PcentOutSegs to get a
        percentage of total rx/tx segments.
        """
        if fld.endswith('PcentInSegs'):
            fld = fld.partition('PcentInSegs')[0]
            return self._percent_in_segs(fld)
        if fld.endswith('PcentOutSegs'):
            fld = fld.partition('PcentOutSegs')[0]
            return self._percent_out_segs(fld)

        return super().__getattr__(fld)


class SockStat(ProcNetBase):
    """
    Provides a common way to extract fields from /proc/net/sockstat.

    Expected file format is:

    hdr1: label1 value1 label2 value2...
    hdr2: label1 value1 label2 value2...
    hdr3: label1 value1 label2 value2...
    """
    def __init__(self):
        super().__init__()
        self._process_file(os.path.join(HotSOSConfig.data_root,
                                        'proc/net/sockstat'))
        # Might happen when sockstat file is not present
        if "TCP" not in self._data:
            self._data["TCP"] = {}
        if "UDP" not in self._data:
            self._data["UDP"] = {}

        self._maybe_parse_sysctl_net_ipv4_xmem("TCP")
        self._maybe_parse_sysctl_net_ipv4_xmem("UDP")

    def _maybe_parse_sysctl_net_ipv4_xmem(self, key):
        xmem = getattr(SYSCtlFactory(), f'net.ipv4.{key.lower()}_mem')
        # Some environments may not have all the sysctl
        # parameters present, so make parsing `optional`
        if xmem is not None:
            (mem_min, mem_pressure, mem_max) = xmem.split(" ")
            self._data[key]["sysctl_mem_min"] = int(mem_min)
            self._data[key]["sysctl_mem_pressure"] = int(mem_pressure)
            self._data[key]["sysctl_mem_max"] = int(mem_max)
            if "mem" in self._data[key]:
                self._data[key]["statistics_mem_usage_pct"] = round(
                    (self._data[key]["mem"] /
                        self._data[key]["sysctl_mem_max"] if
                        self._data[key]["sysctl_mem_max"] else 0) * 100.0, 2)

    def _process_file(self, fname):
        if not os.path.exists(fname):
            log.debug("file not found '%s' - skipping load", fname)
            return None

        log.debug("start processing %s", fname)
        with open(fname, encoding='utf-8') as fd:
            for line in fd:
                psl = line.split(':')
                if len(psl) != 2:
                    log.warning("badness: line does not match format %s", line)
                    continue
                (label, stats) = psl
                if label not in self._data:
                    self._data[label] = {}
                try:
                    stats = stats.strip().split(' ')
                    stats = dict(stats[i:i + 2]
                                 for i in range(0, len(stats), 2))
                    # Convert string values to `int``
                    stats = {k: int(v) for k, v in stats.items()}
                    self._data[label] = stats
                except ValueError as e:
                    log.warning("badness: failed to parse statistics "
                                "for `%s`, bad data `%s` (%s)",
                                label, stats, e)
                    continue

        return None

    @property
    def _header(self):
        # We do not expect this to be used
        raise NotImplementedError()

    @property
    def _fields(self):
        return {
            # Global
            "GlobTcpSocksOrphaned": ("TCP", "orphan"),
            "GlobTcpSocksAllocated": ("TCP", "alloc"),
            "GlobTcpSocksTotalMemPages": ("TCP", "mem"),
            "GlobUdpSocksTotalMemPages": ("UDP", "mem"),
            # Per-netns
            "NsTotalSocksInUse": ("sockets", "used"),
            "NsTcpSocksInUse": ("TCP", "inuse"),
            "NsUdpSocksInUse": ("UDP", "inuse"),
            "NsRawSocksInUse": ("RAW", "inuse"),
            "NsFragSocksInUse": ("FRAG", "inuse"),
            "NsUdpliteSocksInUse": ("UDPLITE", "inuse"),
            "NsTcpSocksInTimeWait": ("TCP", "tw"),
            "NsFragSocksTotalMemPages": ("FRAG", "memory"),
            # These values are from sysctl:
            "SysctlTcpMemMin": ("TCP", "sysctl_mem_min"),
            "SysctlTcpMemPressure": ("TCP", "sysctl_mem_pressure"),
            "SysctlTcpMemMax": ("TCP", "sysctl_mem_max"),
            "SysctlUdpMemMin": ("UDP", "sysctl_mem_min"),
            "SysctlUdpMemPressure": ("UDP", "sysctl_mem_pressure"),
            "SysctlUdpMemMax": ("UDP", "sysctl_mem_max"),
            "UDPMemUsagePct": ("UDP", "statistics_mem_usage_pct"),
            "TCPMemUsagePct": ("TCP", "statistics_mem_usage_pct"),
        }

    def __getattr__(self, fld):
        if fld in self._fields:
            return self._get_field(
                self._fields[fld][0],
                self._fields[fld][1]
            )

        raise AttributeError(f"{fld} not found in {self.__class__.__name__}")


class STOVParserBase(UserList):
    """ Base class for implementations of Standard Table of Values. """
    def __init__(self):
        super().__init__()
        self._load()

    @abc.abstractmethod
    def _load(self):
        """ Load data from source. """

    @property
    @abc.abstractmethod
    def _search_field_info(self):
        """
        Returns a ResultFieldInfo object to make columns addressable by their
        name as well as mapping them to a specific type.
        """

    @property
    @abc.abstractmethod
    def _header_matcher(self):
        """ python.re regular expression to match each header. """

    @property
    @abc.abstractmethod
    def _field_matcher(self):
        """ python.re regular expression to match each row. """

    @property
    @abc.abstractmethod
    def fields(self):
        """
        This is a dictionary of fields exposed by this object along with their
        type. This provides an opportunity to split/translate fields into sub
        fields but can also mirror _search_field_info.
        """


class FieldTranslator():
    """ Field translator. """
    def __init__(self, result, translations):
        self.result = result
        self.translations = translations

    def __getattr__(self, key):
        if key in self.translations:
            return self.translations[key](self.result)

        return getattr(self.result, key)


class Lsof(STOVParserBase):
    """
    Provides a way to extract fields from lsof output.

    Expected file format is:

    COMMAND PID USER  FD TYPE DEVICE SIZE/OFF  NODE NAME
    systemd   1 0 cwd  DIR    9,1     4096     2 /
    systemd   1 0 rtd  DIR    9,1     4096     2 /
    systemd   1 0 txt  REG    9,1  1589552 54275 /lib/
    systemd   1 0 mem  REG    9,1    18976 51133 /lib/
    /*...*/
    """

    def _load(self):
        search = FileSearcher()
        with CLIHelperFile() as cli:
            fout = cli.lsof_Mnlc()
            search.add(SearchDef(self._header_matcher, tag='header'), fout)
            search.add(SearchDef(self._field_matcher, tag='content',
                                 field_info=self._search_field_info), fout)
            results = search.run()
            for r in results.find_by_tag('content'):
                self.data.append(r)

    @staticmethod
    def _int_if_numeric(value):
        if value.isnumeric():
            return int(value)

        return value

    @staticmethod
    def _strip(value):
        return value.strip()

    @property
    def _search_field_info(self):
        finfo = OrderedDict({'COMMAND': str, 'PID': int, 'USER': int,
                             'FD': str, 'TYPE': str, 'DEVICE': str,
                             'SIZE/OFF': str,
                             # NOTE(mkg): Although this column is designated as
                             #            `inode` number, sockets tend to write
                             #            L4 protocol name (e.g. TCP) here as
                             #            well..
                             'NODE': self._int_if_numeric,
                             'NAME': self._strip})
        return ResultFieldInfo(finfo)

    @property
    def _header_matcher(self):
        return r'\s+'.join(self._search_field_info.keys())

    @property
    def _field_matcher(self):
        expr = []
        for key, vtype in self._search_field_info.items():
            if vtype == int:
                _expr = r'(\d+)'
            elif key == 'NAME':
                _expr = r'(\S+\s*\S*)'
            else:
                _expr = r'(\S+)'

            # These fields can be empty/None
            if key in ['TYPE', 'DEVICE', 'SIZE/OFF', 'NODE']:
                _expr = f'{_expr}?'

            expr.append(_expr)

        return r'\s*' + r'\s+'.join(expr)

    @property
    def fields(self):
        """ This has a one-to-one mapping. """
        return self._search_field_info

    def all_with_inode(self, inode):
        return list(filter(lambda x: (x.NODE == inode), self.data))


class NetLink(STOVParserBase):
    """
    Provides a way to extract fields from /proc/net/netlink.

    Expected file format is:

    sk               Eth Pid    Groups   Rmem Wmem Dump Locks Drops Inode
    0000000000000000 0   23984  00000113 0    0    0    2     0     129906
    0000000000000000 0   142171 00000113 0    0    0    2     0     411370
    /*...*/
    """

    def _load(self):
        search = FileSearcher()
        path = os.path.join(HotSOSConfig.data_root, 'proc/net/netlink')
        search.add(SearchDef(self._header_matcher, tag='header'), path)
        search.add(SearchDef(self._field_matcher, tag='content',
                             field_info=self._search_field_info), path)
        results = search.run()
        for r in results.find_by_tag('content'):
            self.data.append(FieldTranslator(r, self.fields))

    @property
    def _search_field_info(self):
        finfo = OrderedDict({'sk': int, 'Eth': int, 'Pid': int,
                             'Groups': str, 'Rmem': int, 'Wmem': int,
                             'Dump': int, 'Locks': int, 'Drops': int,
                             'Inode': int})
        return ResultFieldInfo(finfo)

    @property
    def _header_matcher(self):
        return r'\s+'.join(self._search_field_info.keys())

    @property
    def _field_matcher(self):
        expr = []
        for vtype in self._search_field_info.values():
            if vtype == int:
                _expr = r'(\d+)'
            else:
                _expr = r'(\S+)'

            expr.append(_expr)

        return r'\s*' + r'\s+'.join(expr)

    @property
    def fields(self):
        return {
            # socket pointer addr
            'sk_addr': lambda result: int(str(getattr(result, 'sk')), 16),
            # which protocol this socket belongs in this network family
            'sk_protocol': lambda result: getattr(result, 'sk'),
            # Netlink port id (Pid)
            'netlink_port_id': lambda result: getattr(result, 'Pid'),
            # Netlink groups
            'netlink_groups': lambda result: int(str(getattr(result,
                                                             'Groups')), 16),
            # Allocated rmem for the socket
            'sk_rmem': lambda result: getattr(result, 'Rmem'),
            # Allocated wmem for the socket
            'sk_wmem': lambda result: getattr(result, 'Wmem'),
            # Netlink dump running?
            'netlink_dump': lambda result: getattr(result, 'Dump'),
            # Socket reference count
            'sk_references': lambda result: getattr(result, 'Locks'),
            # Dropped packet counter
            'sk_drops': lambda result: getattr(result, 'Drops'),
            # Socket's inode number
            'sk_inode_num': lambda result: getattr(result, 'Inode'),
        }

    @property
    def all_with_drops(self):
        v = list(filter(lambda x: (x.sk_drops > 0), self.data))
        if v:
            # Correlate netlink sockets with process id's by inode
            # only if there's matching data.
            lsof = Lsof()
            for nlsock in v:
                correlate_result = lsof.all_with_inode(nlsock.sk_inode_num)
                nlsock.procs = set(
                    f"{v.COMMAND}/{v.PID}" for v in correlate_result)
        return v

    @property
    def all_with_drops_str(self):
        if not self.all_with_drops:
            return None

        drops_info = [
            f"inode={nlsock.sk_inode_num} "
            f"procs=[{', '.join(nlsock.procs)}]"
            for nlsock in self.all_with_drops]
        return "\n".join(drops_info)



from collections import UserDict

from hotsos.core.host_helpers.common import get_ps_axo_flags_available
from hotsos.core.host_helpers.cli.common import (
    BinCmd,
    DateFileCmd,
    FileCmd,
)
from hotsos.core.host_helpers.cli.commands import kubectl, ovs, ovn, ceph


class DateBinCmd(BinCmd):
    """ Implementation binary date command. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_hook("pre-exec", self.format_date_cmd)

    def format_date_cmd(self, **kwargs):
        """ Add formatting to date command. """
        no_format = kwargs.get('no_format', False)
        fmt = kwargs.get('format')
        if not no_format and fmt is None:
            fmt = '+%s'

        self.cmd = f'{self.cmd} --utc'
        if fmt:
            # this can't get split() so add to the end of the command list
            self.cmd_extras = [fmt]


class CommandCatalog(UserDict):
    """ Catalog of all supported commands. """

    def __init__(self):
        super().__init__()
        self.data = {
            'apt_config_dump':
                [BinCmd('apt-config dump'),
                 FileCmd('sos_commands/apt/apt-config_dump')],
            'apparmor_status':
                [BinCmd('apparmor_status'),
                 FileCmd('sos_commands/apparmor/apparmor_status')],
            'ceph_daemon_osd_config_show':
                [BinCmd('ceph daemon osd.{osd_id} config show',
                        json_decode=True),
                 # requires sosreport 4.3 or above
                 FileCmd('sos_commands/ceph_osd/'
                         'ceph_daemon_osd.{osd_id}_config_show',
                         json_decode=True)],
            'ceph_daemon_osd_dump_mempools':
                [BinCmd('ceph daemon osd.{osd_id} dump mempools',
                        json_decode=True),
                 # requires sosreport 4.3 or above
                 FileCmd('sos_commands/ceph_osd/'
                         'ceph_daemon_osd.{osd_id}_dump_mempools',
                         json_decode=True)],
            'ceph_health_detail_json_decoded': ceph.CephHealthDetailCommands(),
            'ceph_mon_dump_json_decoded': ceph.CephMonDumpCommands(),
            'ceph_osd_dump_json_decoded': ceph.CephOSDDumpCommands(),
            'ceph_df_json_decoded': ceph.CephDFCommands(),
            'ceph_osd_df_tree_json_decoded': ceph.CepOSDDFTreeCommands(),
            'ceph_osd_crush_dump_json_decoded':
                ceph.CephOSDCrushDumpCommands(),
            'ceph_pg_dump_json_decoded': ceph.CephPGDumpCommands(),
            'ceph_status_json_decoded': ceph.CephStatusCommands(),
            'ceph_versions': ceph.CephVersionsCommands(),
            'ceph_volume_lvm_list': ceph.CephVolumeLVMListCommands(),
            'ceph_report_json_decoded': ceph.CephReportCommands(),
            'ceph_mgr_module_ls': ceph.CephMgrModuleLsCommands(),
            'date':
                [DateBinCmd('date', singleline=True),
                 DateFileCmd('sos_commands/date/date', singleline=True),
                 # this is for legacy sosreport versions
                 DateFileCmd('sos_commands/general/date', singleline=True)],
            'df':
                [BinCmd('df'),
                 FileCmd('df')],
            'dmesg':
                [BinCmd('dmesg'),
                 FileCmd('sos_commands/kernel/dmesg')],
            'docker_images':
                [BinCmd('docker images'),
                 FileCmd('sos_commands/docker/docker_images')],
            'docker_ps':
                [BinCmd('docker ps'),
                 FileCmd('sos_commands/docker/docker_ps')],
            'dpkg_l':
                [BinCmd('dpkg -l'),
                 FileCmd('sos_commands/dpkg/dpkg_-l',
                         decode_error_handling='surrogateecape')],
            'ethtool':
                [BinCmd('ethtool {interface}'),
                 FileCmd('sos_commands/networking/ethtool_{interface}')],
            'hostname':
                [BinCmd('hostname', singleline=True),
                 FileCmd('hostname', singleline=True)],
            'hostnamectl':
                [BinCmd('hostnamectl'),
                 FileCmd('sos_commands/host/hostnamectl_status')],
            'ip_netns':
                [BinCmd('ip netns'),
                 FileCmd('sos_commands/networking/ip_netns')],
            'ip_addr':
                [BinCmd('ip -d address'),
                 FileCmd('sos_commands/networking/ip_-d_address')],
            'ip_link':
                [BinCmd('ip -s -d link'),
                 FileCmd('sos_commands/networking/ip_-s_-d_link')],
            'kubectl_get':
                kubectl.KubectlGetCmds(),
            'kubectl_logs':
                kubectl.KubectlLogsCmds(),
            'ls_lanR_sys_block':
                [BinCmd('ls -lanR /sys/block/'),
                 FileCmd('sos_commands/block/ls_-lanR_.sys.block')],
            'lscpu':
                [BinCmd('lscpu'),
                 FileCmd('sos_commands/processor/lscpu')],
            'lsof_Mnlc':
                [BinCmd('lsof +M -n -l -c ""'),
                 FileCmd('sos_commands/process/lsof_M_-n_-l_-c')],
            'lxd_buginfo':
                [BinCmd('lxd.buginfo'),
                 FileCmd('sos_commands/lxd/lxd.buginfo')],
            'numactl':
                [BinCmd('numactl --hardware'),
                 FileCmd('sos_commands/numa/numactl_--hardware')],
            'ns_ip_addr':
                [BinCmd('ip netns exec {namespace} ip address show'),
                 FileCmd('sos_commands/networking/'
                         'ip_netns_exec_{namespace}_ip_address_show'),
                 FileCmd('sos_commands/networking/namespaces/{namespace}/'
                         'ip_netns_exec_{namespace}_ip_-d_address_show')],
            'ovn_nbctl_list':
                ovn.OVNDBCTLListCmds(ctl_command='ovn-nbctl'),
            'ovn_sbctl_list':
                ovn.OVNDBCTLListCmds(ctl_command='ovn-sbctl'),
            'ovn_nbctl_show':
                ovn.OVNDBCTLShowCmds(ctl_command='ovn-nbctl'),
            'ovn_sbctl_show':
                ovn.OVNDBCTLShowCmds(ctl_command='ovn-sbctl'),
            'ovs_vsctl_get': ovs.OVSVSCtlGetCmds(),
            'ovs_vsctl_list': ovs.OVSVSCtlListCmds(),
            'ovs_vsctl_list_br': ovs.OVSVSCtlListBrCmds(),
            'ovs_appctl': ovs.OVSAppCtlCmds(),
            'ovs_ofctl': ovs.OVSOFCtlCmds(),
            'pacemaker_crm_status':
                [BinCmd('crm status'),
                 FileCmd('sos_commands/pacemaker/crm_status')],
            'pebble_services':
                [BinCmd('pebble services'),
                 # This is how operator charms run it
                 BinCmd('/charm/bin/pebble services'),
                 # The following does not exist in sosreport yet but adding
                 # since it is useful for testing and will hopefully be
                 # supported in sos at some point.
                 FileCmd('sos_commands/pebble/pebble_services')],
            'ps':
                [BinCmd('ps auxwww'),
                 FileCmd('ps')],
            'ps_axo_flags':
                [BinCmd('ps axo flags,state,uid,pid,ppid,pgid,sid,cls,'
                        'pri,addr,sz,wchan:20,lstart,tty,time,cmd'),
                 # Older sosrepot uses 'wchan' option while newer ones use
                 # 'wchan:20' - thus the glob is to cover both
                 FileCmd(get_ps_axo_flags_available() or "")],
            'pro_status':
                [BinCmd('ua status'),
                 FileCmd('sos_commands/ubuntu/ua_status')],
            'rabbitmqctl_report':
                [BinCmd('rabbitmqctl report'),
                 FileCmd('sos_commands/rabbitmq/rabbitmqctl_report')],
            'sunbeam_cluster_list':
                [BinCmd('sunbeam cluster list'),
                 FileCmd('sos_commands/sunbeam/sunbeam_cluster_list')],
            'sunbeam_cluster_list_yaml_decoded':
                [BinCmd('sunbeam cluster list --format yaml'),
                 FileCmd('sos_commands/sunbeam/'
                         'sunbeam_cluster_list_--format_yaml',
                         yaml_decode=True)],
            'snap_list_all':
                [BinCmd('snap list --all'),
                 FileCmd('sos_commands/snap/snap_list_--all'),
                 # sos legacy
                 FileCmd('sos_commands/snappy/snap_list_--all')],
            'sysctl_all':
                [BinCmd('sysctl -a'),
                 FileCmd('sos_commands/kernel/sysctl_-a')],
            'systemctl_status_all':
                [BinCmd('systemctl status --all'),
                 FileCmd('sos_commands/systemd/systemctl_status_--all',
                         decode_error_handling='backslashreplace')],
            'systemctl_list_units':
                [BinCmd('systemctl list-units'),
                 FileCmd('sos_commands/systemd/systemctl_list-units')],
            'systemctl_list_unit_files':
                [BinCmd('systemctl list-unit-files'),
                 FileCmd('sos_commands/systemd/systemctl_list-unit-files')],
            'udevadm_info_dev':
                [BinCmd('udevadm info /dev/{device}'),
                 FileCmd('sos_commands/block/udevadm_info_.dev.{device}')],
            'udevadm_info_exportdb':
                [BinCmd('udevadm info --export-db'),
                 FileCmd('sos_commands/devices/udevadm_info_--export-db')],
            'uname':
                [BinCmd('uname -a', singleline=True),
                 FileCmd('sos_commands/kernel/uname_-a', singleline=True)],
            'uptime':
                [BinCmd('uptime', singleline=True),
                 FileCmd('uptime', singleline=True)],
        }



# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' keystone/exception.py
KEYSTONE_EXCEPTIONS = [
    "_KeystoneExceptionMeta",
    "Error",
    "ValidationError",
    "URLValidationError",
    "PasswordValidationError",
    "PasswordRequirementsValidationError",
    "PasswordHistoryValidationError",
    "PasswordAgeValidationError",
    "PasswordSelfServiceDisabled",
    "SchemaValidationError",
    "ValidationTimeStampError",
    "InvalidOperatorError",
    "ValidationExpirationError",
    "StringLengthExceeded",
    "AmbiguityError",
    "ApplicationCredentialValidationError",
    "CircularRegionHierarchyError",
    "ForbiddenNotSecurity",
    "PasswordVerificationError",
    "RegionDeletionError",
    "ApplicationCredentialLimitExceeded",
    "CredentialLimitExceeded",
    "SecurityError",
    "Unauthorized",
    "InsufficientAuthMethods",
    "ReceiptNotFound",
    "PasswordExpired",
    "AuthPluginException",
    "UserDisabled",
    "AccountLocked",
    "AuthMethodNotSupported",
    "ApplicationCredentialAuthError",
    "AdditionalAuthRequired",
    "Forbidden",
    "ForbiddenAction",
    "CrossBackendNotAllowed",
    "InvalidPolicyAssociation",
    "InvalidDomainConfig",
    "InvalidLimit",
    "LimitTreeExceedError",
    "NotFound",
    "EndpointNotFound",
    "PolicyNotFound",
    "PolicyAssociationNotFound",
    "RoleNotFound",
    "ImpliedRoleNotFound",
    "InvalidImpliedRole",
    "DomainSpecificRoleMismatch",
    "DomainSpecificRoleNotWithinIdPDomain",
    "DomainIdInvalid",
    "RoleAssignmentNotFound",
    "RegionNotFound",
    "ServiceNotFound",
    "DomainNotFound",
    "ProjectNotFound",
    "ProjectTagNotFound",
    "TokenNotFound",
    "UserNotFound",
    "GroupNotFound",
    "MappingNotFound",
    "TrustNotFound",
    "TrustUseLimitReached",
    "CredentialNotFound",
    "VersionNotFound",
    "EndpointGroupNotFound",
    "IdentityProviderNotFound",
    "ServiceProviderNotFound",
    "FederatedProtocolNotFound",
    "PublicIDNotFound",
    "RegisteredLimitNotFound",
    "LimitNotFound",
    "NoLimitReference",
    "RegisteredLimitError",
    "DomainConfigNotFound",
    "ConfigRegistrationNotFound",
    "ApplicationCredentialNotFound",
    "AccessRuleNotFound",
    "Conflict",
    "UnexpectedError",
    "TrustConsumeMaximumAttempt",
    "MalformedEndpoint",
    "MappedGroupNotFound",
    "MetadataFileError",
    "DirectMappingError",
    "AssignmentTypeCalculationError",
    "NotImplemented",
    "Gone",
    "ConfigFileNotFound",
    "KeysNotFound",
    "MultipleSQLDriversInConfig",
    "MigrationNotProvided",
    "UnsupportedTokenVersionException",
    "SAMLSigningError",
    "OAuthHeadersMissingError",
    "TokenlessAuthConfigError",
    "CredentialEncryptionError",
    "LDAPServerConnectionError",
    "LDAPInvalidCredentialsError",
    "LDAPSizeLimitExceeded",
    "CacheDeserializationError",
    "ResourceUpdateForbidden",
    "ResourceDeleteForbidden",
]



from hotsos.core.plugins.storage.bcache import BcacheChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class BcacheSummary(BcacheChecks):
    """ Implementation of Bcache summary. """
    summary_part_index = 2

    # REMINDER: Common entries are implemented in
    #           plugintools.ApplicationSummaryBase. Only customisations are
    #           implemented here. See
    #           plugintools.get_min_available_entry_index() for an explanation
    #           on how entry indices are managed.

    @summary_entry('cachesets', get_min_available_entry_index())
    def summary_cachesets(self):
        _state = {}
        for cset in self.cachesets:
            _cset = {'cache_available_percent':
                     int(cset.cache_available_percent), 'bdevs': {}}
            for bdev in cset.bdevs:
                _cset['bdevs'][bdev.name] = {'dev': bdev.dev,
                                             'backing_dev':
                                             bdev.backing_dev_name}
                if bdev.dev_to_dname:
                    _cset['bdevs'][bdev.name]['dname'] = bdev.dev_to_dname

            _state[cset.uuid] = _cset

        return _state



from collections import UserList

from hotsos.core.host_helpers.cli.common import BinCmd, FileCmd

KubectlAliases = ['kubectl', 'microk8s.kubectl', 'k8s kubectl']

# Paths checked by sos report
KUBE_CONFIGS = ['/etc/kubernetes/admin.conf',
                '/root/cdk/cdk_addons_kubectl_config',
                '/var/snap/microk8s/current/credentials/client.config']


class KubectlBinCmdBase(BinCmd):
    """ Generic kubectl implementation for binary kubectl commands.

    Aims to support all kubectl commands and their subcommands in a format
    comparable to sosreport. This means that if e.g. sosreport executes the
    command with -o json then we need to do the same here for the binary
    version.
    """
    KUBECTL_CMD = None

    @property
    def _use_json_encoding(self):
        return self.KUBECTL_CMD == 'get'

    def __init__(self, cmd, *args, **kwargs):
        cmd = (f'{cmd} {self.KUBECTL_CMD} --namespace {{namespace}} {{opt}} '
               '{subopts}')

        # NOTE: sosreport runs kubectl get with -o json so we will do the
        # same for binary commands so that the api calls can be expect
        # consistent return type.
        if self._use_json_encoding:
            kwargs['json_decode'] = True

        super().__init__(cmd, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self._use_json_encoding:
            _subopts = kwargs.get('subopts', "")
            if not _subopts:
                kwargs['subopts'] = "-o json"
            else:
                kwargs['subopts'] += " -o json"

        return super().__call__(*args, **kwargs)


class KubectlLogsBinCmd(KubectlBinCmdBase):
    """ Implements kubectl logs binary command. """
    KUBECTL_CMD = 'logs'


class KubectlFileBinCmd(FileCmd):
    """ Implements kubectl logs file command. """
    KUBECTL_CMD = 'logs'

    def __call__(self, *args, **kwargs):
        subopts = kwargs.get('subopts')
        if subopts:
            kwargs['subopts'] = subopts.replace(' ', '_')

        return super().__call__(*args, **kwargs)


class KubectlLogsCmds(UserList):
    """ Generate kubectl logs command variants. """

    def __init__(self):
        cmds = [KubectlLogsBinCmd(cmd) for cmd in KubectlAliases]

        confs = [p.replace('/', '.') for p in KUBE_CONFIGS]
        paths = []
        for cmd in KubectlAliases:
            cmd = cmd.replace('.', '_').replace(' ', '_')
            paths.append('sos_commands/kubernetes/cluster-info/{namespace}/'
                         'podlogs/{opt}/' + cmd +
                         '_--namespace_{namespace}_logs_{opt}_{subopts}')

            paths.extend([(f'sos_commands/kubernetes/pods/{cmd}'
                           f'_--kubeconfig_{conf}'
                           '_--namespace_{namespace}_logs_{opt}'
                           '_{subopts}') for conf in
                          confs])

        for path in paths:
            cmds.append(KubectlFileBinCmd(path))

        super().__init__(cmds)


class KubectlGetBinCmd(KubectlBinCmdBase):
    """ Implements kubectl get """
    KUBECTL_CMD = 'get'


class KubectlGetCmds(UserList):
    """ Generate kubectl get command variants. """

    def __init__(self):
        cmds = [KubectlGetBinCmd(cmd) for cmd in KubectlAliases]

        confs = [p.replace('/', '.') for p in KUBE_CONFIGS]
        paths = []
        for cmd in KubectlAliases:
            cmd = cmd.replace('.', '_').replace(' ', '_')
            paths.extend([(f'sos_commands/kubernetes/services/{cmd}'
                           f'_--kubeconfig_{conf}'
                           '_get_-o_json_--namespace_{namespace}_{opt}')
                          for conf in confs])

            paths.append('sos_commands/kubernetes/cluster-info/{namespace}/'
                         + cmd +
                         '_get_-o_json_--namespace_{namespace}_{opt}')

        for path in paths:
            cmds.append(FileCmd(path, json_decode=True))

        super().__init__(cmds)



from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.common import YPropertyOverrideBase


class YPropertyVars(YPropertyOverrideBase):
    """
    Vars property. Provides support for defining variables in YAML e.g.

    vars:
      var1: val1
      var2: @property1

    where val1 is a literal and property2 is an import path.
    """
    override_keys = ['vars']
    _allow_subtree = False

    @staticmethod
    def _is_import_path(value):
        return isinstance(value, str) and value.startswith('@')

    def resolve(self, name):
        """
        Value can be a raw type or a string starting with a '@' which indicates
        it it a Python property to be imported/loaded. Imports are done when
        the variable value is accessed for the first time i.e. it is lazy
        loaded.
        """
        log.debug("resolving var '%s'", name)
        if not name:
            raise NameError(
                f"attempting to resolve invalid varname '{name}'")

        if hasattr(self, name):
            value = getattr(self, name)
            if self._is_import_path(value):
                # NOTE: this path can be a property or a factory.
                value = self.get_property(self.context, value[1:])

            log.debug("resolved var %s=%s (type=%s)", name, value, type(value))
        else:
            log.warning("could not resolve var %s", name)
            value = None

        return value

    def __iter__(self):
        log.debug("iterating over vars")
        for name, var in self._vardefs.items():
            log.debug("returning var %s", name)
            yield name, var.vardef



import re

from hotsos.core.plugins.kernel.common import KernelChecks
from hotsos.core.plugins.kernel.config import SystemdConfig
from hotsos.core.plugins.kernel.memory import MemoryChecks
from hotsos.core.plugins.kernel.sysfs import CPU
from hotsos.core.plugintools import summary_entry


class KernelSummary(KernelChecks):
    """ Implementation of Kernel summary. """
    summary_part_index = 0

    # REMINDER: Common entries are implemented in
    #           plugintools.ApplicationSummaryBase. Only customisations are
    #           implemented here. See
    #           plugintools.get_min_available_entry_index() for an explanation
    #           on how entry indices are managed.

    @property
    def cpu_info(self):
        cpu = CPU()
        info = {}
        if cpu.vendor:
            info['vendor'] = cpu.vendor

        if cpu.model:
            model = cpu.model.lower()
            ret = None
            # strip intel or amd trailing info
            if 'intel' in cpu.vendor:
                ret = re.search(r'(.+) cpu @ .+', model)
            elif 'amd' in cpu.vendor:
                ret = re.search(r'(.+) \d+-core .+', model)

            if ret:
                model = ret.group(1).strip()

            info['model'] = model

        if cpu.smt is not None:
            if cpu.smt:
                info['smt'] = 'enabled'
            else:
                info['smt'] = 'disabled'

        if cpu.isolated is not None and cpu.isolated != '':
            info['isolated'] = cpu.isolated

        cpu_gov_all = cpu.cpufreq_scaling_governor_all
        info['cpufreq-scaling-governor'] = cpu_gov_all
        return info

    @summary_entry("version", 0)
    def summary_version(self):
        return self.version or None

    @summary_entry("boot", 1)
    def summary_boot(self):
        if self.boot_parameters:
            return ' '.join(self.boot_parameters)

        return None

    @staticmethod
    @summary_entry("systemd", 2)
    def summary_systemd():
        cfg = SystemdConfig()
        if cfg.exists:
            if cfg.get('CPUAffinity'):
                return {'CPUAffinity': cfg.get('CPUAffinity')}

        return None

    @summary_entry("cpu", 3)
    def summary_cpu(self):
        return self.cpu_info or None

    @staticmethod
    @summary_entry("memory", 4)
    def summary_memory():
        return MemoryChecks().nodes_with_limited_high_order_memory_full or None



from .common import (
    CephChecks,
    CephConfig,
    CephDaemonAllOSDsFactory,
    PathFinder,
)
from .cluster import CephCluster, CephCrushMap

__all__ = [
    CephChecks.__name__,
    CephConfig.__name__,
    CephDaemonAllOSDsFactory.__name__,
    CephCluster.__name__,
    CephCrushMap.__name__,
    PathFinder.__name__,
    ]



import re
import subprocess
from functools import cached_property

from hotsos.core.host_helpers.common import get_ps_axo_flags_available
from hotsos.core.host_helpers import (
    CLIHelper,
    CLIHelperFile,
    SystemdHelper,
)
from hotsos.core.search import (
    FileSearcher,
    SearchDef
)
from hotsos.core.utils import seconds_to_date


class CephDaemonBase():
    """ Base class for all Ceph daemon implementations. """
    def __init__(self, daemon_type):
        self.daemon_type = daemon_type
        self.id = None
        self.date_in_secs = self.get_date_secs()

    @classmethod
    def get_date_secs(cls, datestring=None):
        if datestring:
            cmd = ["date", "--utc", f"--date={datestring}", "+%s"]
            date_in_secs = subprocess.check_output(cmd)
        else:
            date_in_secs = CLIHelper().date() or 0
            if date_in_secs:
                date_in_secs = date_in_secs.strip()

        return int(date_in_secs)

    @cached_property
    def rss(self):
        """Return memory RSS for a given daemon.

        NOTE: this assumes we have ps auxwwwm format.
        """
        s = FileSearcher()
        # columns: USER PID %CPU %MEM VSZ RSS TTY STAT START TIME COMMAND
        if self.id is not None:
            ceph_id = rf"--id\s+{self.id}"
        else:
            ceph_id = ''

        expr = (r"\S+\s+\d+\s+\S+\s+\S+\s+\d+\s+(\d+)\s+.+/ceph-"
                rf"{self.daemon_type}\s+.+{ceph_id}\s+.+.+")
        sd = SearchDef(expr)
        with CLIHelperFile() as cli:
            ps_out = cli.ps()
            s.add(sd, path=ps_out)
            rss = 0
            # we only expect one result
            for result in s.run().find_by_path(ps_out):
                rss = int(int(result.get(1)) / 1024)
                break

        return f"{rss}M"

    @cached_property
    def etime(self):
        """Return process etime for a given daemon.

        To get etime we have to use ps_axo_flags rather than the default
        ps_auxww.
        """
        if not get_ps_axo_flags_available():
            return None

        if self.id is None:
            return None

        ps_info = []
        daemon = f"ceph-{self.daemon_type}"
        for line in CLIHelper().ps_axo_flags():
            ret = re.compile(daemon).search(line)
            if not ret:
                continue

            expt_tmplt = SystemdHelper.PS_CMD_EXPR_TEMPLATES['absolute']
            ret = re.compile(expt_tmplt.format(daemon)).search(line)
            if ret:
                ps_info.append(ret.group(0))

        if not ps_info:
            return None

        _etime = None
        for cmd in ps_info:
            ret = re.compile(rf".+\s+.+--id {self.id}\s+.+").match(cmd)
            if ret:
                osd_start = ' '.join(cmd.split()[13:17])
                if self.date_in_secs and osd_start:
                    osd_start_secs = self.get_date_secs(datestring=osd_start)
                    osd_uptime_secs = self.date_in_secs - osd_start_secs
                    osd_uptime_str = seconds_to_date(osd_uptime_secs)
                    _etime = osd_uptime_str

        return _etime


class CephMon(CephDaemonBase):
    """ Representation of a Ceph Mon """
    def __init__(self, name):
        super().__init__('mon')
        self.name = name


class CephMDS(CephDaemonBase):
    """ Representation of a Ceph MDS """
    def __init__(self):
        super().__init__('mds')


class CephRGW(CephDaemonBase):
    """ Representation of a Ceph RGW """
    def __init__(self):
        super().__init__('radosgw')


class CephOSD(CephDaemonBase):
    """ Representation of a Ceph OSD """
    def __init__(self, ceph_id, fsid=None, device=None, dump=None):
        super().__init__('osd')
        self.id = ceph_id
        self.fsid = fsid
        self.device = device
        self.dump = dump

    def to_dict(self):
        d = {self.id: {
             'fsid': self.fsid,
             'dev': self.device}}

        if self.devtype:
            d[self.id]['devtype'] = self.devtype

        if self.etime:
            d[self.id]['etime'] = self.etime

        if self.rss:
            d[self.id]['rss'] = self.rss

        return d

    @cached_property
    def devtype(self):
        osd_tree = CLIHelper().ceph_osd_df_tree_json_decoded()
        if not osd_tree:
            return None

        _devtype = None
        for node in osd_tree.get('nodes'):
            if node.get('type') == 'osd' and node['id'] == self.id:
                _devtype = node['device_class']

        return _devtype



from hotsos.core.plugins.lxd import LXD, LXDChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class LXDSummary(LXDChecks):
    """ Implementation of LXD summary. """
    summary_part_index = 0

    # REMINDER: Common entries are implemented in
    #           plugintools.ApplicationSummaryBase. Only customisations are
    #           implemented here. See
    #           plugintools.get_min_available_entry_index() for an explanation
    #           on how entry indices are managed.

    @staticmethod
    @summary_entry('instances', get_min_available_entry_index())
    def summary_instances():
        return LXD().instances or None



from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.issues import IssuesManager, HotSOSScenariosWarning
from hotsos.core.log import log
from hotsos.core.ycheck.engine import (
    YDefsLoader,
    YDefsSection,
    YHandlerBase,
)
from hotsos.core.ycheck.engine.properties.common import YDefsContext
from hotsos.core.ycheck.common import GlobalSearcherPreloaderBase
from hotsos.core.exceptions import AlreadyLoadedError


class ScenariosSearchPreloader(YHandlerBase, GlobalSearcherPreloaderBase):
    """
    Find all scenario checks that have a search property and load their search
    into the global searcher.

    @param global_searcher: GlobalSearcher object
    """

    @property
    def filter(self):
        return HotSOSConfig.scenario_filter

    @cached_property
    def root(self):
        """ Return the root of all scenarios. """
        plugin_defs = YDefsLoader('scenarios').plugin_defs
        return YDefsSection(HotSOSConfig.plugin_name, plugin_defs or {})

    @property
    def scenarios(self):
        """
        Generator for all scenarios.

        Searches are expected to exist within checks so we extract them from
        each scenario and discover their searches.

        @return: scenario
        """
        for scenario in self.root.leaf_sections:
            # NOTE: scenario is not an override object, it is a PTreeSection
            # that contains the objects i.e. we go looking for the checks etc
            # within this scenario (sub)section.
            if self.skip_filtered(self.filter, scenario.resolve_path):
                log.info("skipping scenario %s from pre-search "
                         "(filter=%s)", scenario.resolve_path, self.filter)
                continue

            yield scenario

    def _preload_searches(self):
        """
        Find all scenario checks that have a search property and load their
        search into the global searcher.
        """
        if self.global_searcher.is_loaded(self.__class__.__name__):
            raise AlreadyLoadedError(
                "scenario searches have already been loaded into "
                "the global searcher. This operation can only be "
                "performed once.")

        log.debug("started loading scenario searches into searcher "
                  "(%s)", self.global_searcher.searcher)
        added = 0
        for scenario in self.scenarios:
            checks = scenario.checks
            log.debug("looking for searches in %s", checks.override_path)
            # this is needed to resolve search expression references
            checks.initialise(scenario.vars)
            for check in checks:
                if check.search is None:
                    continue

                # Checks can use input outside their scope so we
                # take that into account here.
                _input = check.input
                if _input is None:
                    _input = scenario.input

                log.debug("loading search for check '%s'", check.name)
                added += 1
                self._load_item_search(self.global_searcher, check.search,
                                       _input)

        self.global_searcher.set_loaded(self.__class__.__name__)
        log.debug("identified a total of %s scenario check searches", added)
        log.debug("finished loading scenario searches into searcher "
                  "(registry now has %s items)", len(self.global_searcher))

    def run(self):
        # Pre-load all scenario check searches into a global searcher
        self._preload_searches()


class Scenario():
    """
    Representation of a scenario containing all checks and conclusions.
    """
    def __init__(self, name, _checks, conclusions):
        log.debug("scenario: %s", name)
        self.name = name
        self._checks = _checks
        self._conclusions = conclusions

    @property
    def checks(self):
        return {c.name: c for c in self._checks}

    @property
    def conclusions(self):
        return {c.name: c for c in self._conclusions}


class YScenarioChecker(YHandlerBase):
    """
    This class handles the loading, execution and processing of all scenarios
    for the current plugin.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._scenarios = []
        hours = 24
        if HotSOSConfig.use_all_logs:
            hours *= HotSOSConfig.max_logrotate_depth

        # It is assumed that the global searcher already exists, is loaded with
        # searches and they have been executed. Unit tests however, should be
        # resetting the registry prior to each run and we will therefore need
        # to load searches each time which is why we do this here. This is
        # therefore not intended to be used outside of a test scenario.
        label = ScenariosSearchPreloader.__name__
        if not self.global_searcher.is_loaded(label):
            log.info("global searcher catalog is empty so launching "
                     "pre-load of scenario searches")
            # NOTE: this is not re-entrant safe and is only ever expected
            #       to be done from a unit test.
            ScenariosSearchPreloader(self.global_searcher).run()

    @property
    def filter(self):
        return HotSOSConfig.scenario_filter

    def load(self):
        plugin_content = YDefsLoader('scenarios').plugin_defs
        if not plugin_content:
            return

        yscenarios = YDefsSection(HotSOSConfig.plugin_name, plugin_content,
                                  context=YDefsContext({'global_searcher':
                                                        self.global_searcher}))
        if (not HotSOSConfig.force_mode and yscenarios.requires and not
                yscenarios.requires.result):
            log.debug("plugin '%s' scenarios pre-requisites not met - "
                      "skipping", HotSOSConfig.plugin_name)
            return

        log.debug("sections=%s, scenarios=%s",
                  len(list(yscenarios.branch_sections)),
                  len(list(yscenarios.leaf_sections)))

        to_skip = set()
        for scenario in yscenarios.leaf_sections:
            if ScenariosSearchPreloader.skip_filtered(self.filter,
                                                      scenario.resolve_path):
                log.info("skipping scenario %s (filter=%s)",
                         scenario.resolve_path,
                         HotSOSConfig.scenario_filter)
                continue

            # Only register scenarios if requirements are satisfied.
            group_name = scenario.parent.name
            if (not HotSOSConfig.force_mode and
                    (group_name in to_skip or
                        (scenario.requires and not scenario.requires.result))):
                log.debug("%s requirements not met - skipping scenario %s",
                          group_name, scenario.name)
                to_skip.add(group_name)
                continue

            scenario.checks.initialise(scenario.vars)
            scenario.checks.check_context.global_searcher = \
                self.global_searcher
            _scenario = Scenario(scenario.name, scenario.checks,
                                 scenario.conclusions)
            scenario.conclusions.initialise(scenario.vars, _scenario.checks)
            self._scenarios.append(_scenario)

    @property
    def scenarios(self):
        return self._scenarios

    @staticmethod
    def _run_scenario_conclusion(scenario, issue_mgr):
        """ Determine the conclusion of this scenario. """
        results = {}
        # Run conclusions in order of priority. One or more conclusion may
        # share the same priority. If one or more conclusion of the same
        # priority is reached the rest (of lower priority) are ignored.
        last_priority = None
        for conc in sorted(scenario.conclusions.values(), key=lambda _conc:
                           int(_conc.priority or 1), reverse=True):
            priority = int(conc.priority or 1)
            if last_priority is not None:
                if priority < last_priority and last_priority in results:
                    break

            last_priority = priority
            if conc.reached(scenario.checks):
                if priority in results:
                    results[priority].append(conc)
                else:
                    results[priority] = [conc]

                log.debug("conclusion reached: %s (priority=%s)", conc.name,
                          priority)

        if results:
            highest = max(results.keys())
            log.debug("selecting highest priority=%s conclusions (%s)",
                      highest, len(results[highest]))
            for conc in results[highest]:
                issue_mgr.add(conc.issue, context=conc.issue_context)
        else:
            log.debug("no conclusions reached")

    def run(self, load=True):
        if load:
            self.load()

        failed_scenarios = []
        issue_mgr = IssuesManager()
        for scenario in self.scenarios:
            log.debug("running scenario: %s", scenario.name)
            # catch failed scenarios and allow others to run
            try:
                self._run_scenario_conclusion(scenario, issue_mgr)
            # We really do want to catch all here since we don't care why
            # it failed but don't want to fail hard if it does.
            except Exception:  # pylint: disable=W0718
                log.exception("caught exception when running scenario %s:",
                              scenario.name)
                failed_scenarios.append(scenario.name)

        if failed_scenarios:
            msg = ("One or more scenarios failed to run "
                   f"({', '.join(failed_scenarios)}) - run hotsos in "
                   "debug mode (--debug) to get more detail")
            issue_mgr.add(HotSOSScenariosWarning(msg))



""" OpenStack Sunbeam helper code. """
from collections import defaultdict
from functools import cached_property

from hotsos.core.host_helpers import CLIHelper, SnapPackageHelper
from hotsos.core.log import log


class SunbeamInfo():
    """ OpenStack Sunbeam helpers. """

    @cached_property
    def is_controller(self):
        return 'openstack' in SnapPackageHelper(core_snaps=['openstack']).core

    @cached_property
    def pods(self):
        """ Return dictionary of pods keyed by their status e.g. Running. """
        if not self.is_controller:
            return {}

        cli = CLIHelper()
        out = cli.kubectl_get(namespace='openstack', opt='pods', subopts='')
        if 'items' not in out:
            log.info('no sunbeam pods found')
            return {}

        pods = defaultdict(list)
        for pod in out['items']:
            phase = pod['status']['phase']
            pods[phase].append(pod['metadata']['name'])

        # must be type dict for pyyaml
        return dict(pods)

    @cached_property
    def statefulsets(self):
        """ Return dictionary of statefulsets grouped under either 'complete'
        or 'incomplete' depending on whether all members are ready.
        """
        if not self.is_controller:
            return {}

        cli = CLIHelper()
        out = cli.kubectl_get(namespace='openstack', opt='statefulsets',
                              subopts='')

        ss = {'complete': [], 'incomplete': []}
        if 'items' not in out:
            log.info('no sunbeam statefulsets found')
            return {}

        metadata_fails = 0
        for i in out['items']:
            try:
                name = i['metadata']['name']
            except KeyError:
                metadata_fails += 1
                continue

            try:
                ready = i['status'].get('readyReplicas', 0)
                if ready and i['status']['replicas'] / ready:
                    ss['complete'].append(name)
                else:
                    ss['incomplete'].append(name)
            except KeyError:
                log.warning("failed to get replicaset '%s' status", name)

        if metadata_fails:
            log.warning("failed to get replicaset metadata for %s items",
                        metadata_fails)

        # must be type dict for pyyaml
        return dict(ss)



import abc
from functools import cached_property
from dataclasses import dataclass

from hotsos.core.log import log
from hotsos.core.search import (
    SearchDef,
    SequenceSearchDef,
    FileSearcher,
)
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelper
from hotsos.core.utils import mktemp_dump
from hotsos.core.plugins.openvswitch.common import (
    OpenvSwitchInstallInfo,
    OVSDBTableBase
)


class OVNDBTable(OVSDBTableBase):
    """
    Provides an interface to an OVNDB table. Records can be extracted from
    either 'get' or 'list' command outputs. We try 'get' first and of not found
    we search in output of 'list'.
    """

    def __init__(self, ovndb_cli_command, *args, **kwargs):
        self._ovndb_cli_command = ovndb_cli_command
        super().__init__(*args, **kwargs)

    @staticmethod
    def _get_cmd(table):
        return lambda **kwargs: None

    def _list_cmd(self, table):  # pylint: disable=arguments-differ
        return lambda **kwargs: getattr(CLIHelper(),
                                        self._ovndb_cli_command)(table=table,
                                                                 **kwargs)


@dataclass
class OVNSBDBPort:
    """ Representation of OVS SBDB port """

    name: str
    port_type: str


class OVNSBDBChassis():
    """ Representation of OVS SBDB chassis """
    def __init__(self, name, content):
        self.name = name
        self.content = content or {}

    @cached_property
    def _ports(self):
        ports = []
        if not self.content or 'Port_Binding' not in self.content:
            return ports

        for p in self.content['Port_Binding']:
            if 'cr-lrp' in p:
                ports.append(OVNSBDBPort(p, 'cr-lrp'))
            else:
                ports.append(OVNSBDBPort(p, 'lsp'))

        return ports

    @cached_property
    def ports(self):
        return self._ports

    @cached_property
    def cr_ports(self):
        return [p for p in self._ports if p.port_type == 'cr-lrp']


class OVNDBBase():
    """ Base class for OVN database objects. """
    def __init__(self):
        contents = mktemp_dump(''.join(self.db_show))
        s = FileSearcher()
        s.add(self.resources_sd, contents)
        self.results = s.run()

    @property
    @abc.abstractmethod
    def db_show(self):
        """ e.g. ovn-sbctl show  """

    @cached_property
    def resources_sd(self):
        start = SearchDef([r"^(\S+)\s+(\S+)\s*.*"])
        body = SearchDef(r"^\s+(\S+)\s+(.+)")
        return SequenceSearchDef(start=start, body=body, tag='resources')

    @cached_property
    def resources(self):
        _resources = {}
        resource = None
        rid = None
        sections = self.results.find_sequence_sections(self.resources_sd)
        for section in sections.values():
            for result in section:
                if result.tag == self.resources_sd.start_tag:
                    resource = result.get(1)
                    rid = result.get(2)
                    if resource in _resources:
                        _resources[resource][rid] = {}
                    else:
                        _resources[resource] = {rid: {}}
                elif result.tag == self.resources_sd.body_tag:
                    rtype = result.get(1)
                    if rtype in _resources[resource][rid]:
                        _resources[resource][rid][rtype].append(result.get(2))
                    else:
                        _resources[resource][rid][rtype] = [result.get(2)]

        return _resources


class OVNNBDB(OVNDBBase):
    """ Represents OVN Northbound DB. """

    @property
    def db_show(self):
        return CLIHelper().ovn_nbctl_show()

    @cached_property
    def routers(self):
        _routers = []
        for r, v in self.resources.items():
            if r == 'router':
                for rid in v:
                    _routers.append(rid)

        return _routers

    @cached_property
    def switches(self):
        _switches = []
        for r, v in self.resources.items():
            if r == 'switch':
                for sid in v:
                    _switches.append(sid)

        return _switches

    def __getattr__(self, table):
        return OVNDBTable('ovn_nbctl_list', table)


class OVNSBDB(OVNDBBase):
    """ Represents OVN Southbound DB. """

    @property
    def db_show(self):
        return CLIHelper().ovn_sbctl_show()

    @cached_property
    def chassis(self):
        _chassis = []
        for r, v in self.resources.items():
            if r == 'Chassis':
                for cid, contents in v.items():
                    _chassis.append(OVNSBDBChassis(cid, contents))

        return _chassis

    def __getattr__(self, table):
        return OVNDBTable('ovn_sbctl_list', table)


class OVNBase():
    """ Base class for all OVN components. """

    @staticmethod
    def _is_host_type(hosttype, check_snap=False):
        if HotSOSConfig.force_mode:
            return True

        if check_snap:
            ret = hosttype in OpenvSwitchInstallInfo().snaps.core
        else:
            ret = hosttype in OpenvSwitchInstallInfo().systemd.services

        if ret:
            log.debug("this host has type={hosttype}")
        else:
            log.debug("this does not have type={hosttype}")

        return ret

    @cached_property
    def is_microovn(self):
        return self._is_host_type('microovn', check_snap=True)

    @cached_property
    def is_ovn_central(self):
        return self._is_host_type('ovn-central')

    @cached_property
    def is_ovn_controller(self):
        return self._is_host_type('ovn-controller')

    @cached_property
    def sbdb(self):
        if self.is_ovn_central or self.is_microovn:
            return OVNSBDB()

        return None

    @cached_property
    def nbdb(self):
        if self.is_ovn_central or self.is_microovn:
            return OVNNBDB()

        return None



from functools import cached_property

from hotsos.core.log import log
from hotsos.core.host_helpers import (
    DPKGVersion
)
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    YRequirementTypeBase,
    PackageCheckItemsBase
)


class BinCheckItems(PackageCheckItemsBase):
    """
    By default we use the APT version checks logic for binary versions. At some
    point in the future we may want to support overriding this.
    """

    def __init__(self, *args, bin_handler=None, **kwargs):
        self.bin_handler = bin_handler
        super().__init__(*args, **kwargs)

    @cached_property
    def packaging_helper(self):
        return self.bin_handler()

    @cached_property
    def installed_versions(self):
        _versions = []
        for p in self.installed:
            _versions.append(self.packaging_helper.get_version(p))

        return _versions


class YRequirementTypeBinary(YRequirementTypeBase):
    """
    Binary requires type property. Provides support for defining binary
    checks.
    """
    override_keys = ['binary']
    override_autoregister = True

    @property
    def _bin_info(self):
        if hasattr(self, 'name'):
            return {self.name: None}

        return {k: v for k, v in self.content.items() if k not in ['handler']}

    @property
    @intercept_exception
    def _result(self):
        items = BinCheckItems(self._bin_info,
                              bin_handler=self.get_cls(self.context,
                                                       self.handler))

        _result = True
        # bail on first fail i.e. if any not installed
        if not items.not_installed:
            for binary, versions in items:
                log.debug("binary %s installed=%s", binary, _result)
                if not versions:
                    continue
                version = items.packaging_helper.get_version(binary)
                _result = DPKGVersion.is_version_within_ranges(version,
                                                               versions)
                # bail at first failure
                if not _result:
                    break
        else:
            log.debug("one or more binary not installed so returning False "
                      "- %s", ', '.join(items.not_installed))
            # bail on first fail i.e. if any not installed
            _result = False

        self.cache.set('binary', ', '.join(items.installed))
        self.cache.set('version', ', '.join([
            str(x) for x in items.installed_versions]))
        log.debug('requirement check: %s %s (result=%s)',
                  self.__class__.__name__,
                  ', '.join(items.packages_to_check), _result)
        return _result



#!/usr/bin/env python3
import contextlib
import os
import subprocess
import sys
import threading
import warnings
from importlib import metadata, resources
from dataclasses import dataclass, fields

import click
import distro
from progress.spinner import Spinner
from hotsos.core import plugintools
from hotsos.core.root_manager import DataRootManager
from hotsos.core.config import HotSOSConfig
from hotsos.core.log import log, LoggingManager
from hotsos.client import (
    HotSOSClient,
    SUPPORTED_SUMMARY_FORMATS
)

SNAP_ERROR_MSG = """ERROR: hotsos is installed as a snap which only supports
running against a sosreport due to access restrictions.

If you want to analyse a host you need to use an alternative installation
method e.g. debian package - see https://hotsos.readthedocs.io/en/latest/install/index.html for more information."
""" # noqa


def is_snap():
    """
    If hotsos is being run as a snap the SNAP_NAME env var will be set to this
    value.
    """
    return os.environ.get('SNAP_NAME', '') == 'hotsos'


def get_os_id():
    """ Return the distro name. """
    return distro.id()


def get_os_version():
    """ Return the distro version. """
    return float(distro.version())


def get_hotsos_root():
    """ Return the hotsos package root path based on the cmd line. """
    return os.path.dirname(sys.argv[0])


def get_version():
    """
    Get hotsos version.

    If run as snap this will come from the SNAP_REVISION env var. Other get
    version from python package.
    """
    ver = os.environ.get('SNAP_REVISION')
    if ver is None:
        try:
            ver = metadata.version('hotsos')
        except metadata.PackageNotFoundError:
            pass

    if ver is not None:
        return ver

    return 'development'


def get_repo_info():
    """
    Get repo-info

    This is generated by whatever package manager was used to build hotsos.
    Only available for built versions.
    """
    repo_info = os.environ.get('REPO_INFO_PATH')
    if repo_info and os.path.exists(repo_info):
        with open(repo_info, encoding='utf-8') as fd:
            return fd.read().strip()

    # NOTE: The pylint warning is suppressed for W4902 because the
    # alternative (i.e. resources.files) is not available for python
    # 3.8, which is a supported environment for hotsos.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        # pylint: disable-next=W4902
        with resources.path('hotsos', '.repo-info') as repo_info:
            if repo_info and os.path.exists(repo_info):
                with open(repo_info, encoding='utf-8') as fd:
                    return fd.read().strip()

    try:
        out = subprocess.check_output(['git', '-C', get_hotsos_root(),
                                       'rev-parse', '--short', 'HEAD'],
                                      stderr=subprocess.DEVNULL)
        if not out:
            return "unknown"
    # We really do want to catch all here since we don't care why it failed
    # but don't want to fail hard if it does.
    except Exception:  # pylint: disable=W0718
        return "unknown"

    return out.decode().strip()


def set_plugin_options(f):
    """ Create a cli option for each plugin. """

    for plugin in plugintools.PLUGINS:
        click.option(f'--{plugin}', default=False, is_flag=True,
                     help=f'Run the {plugin} plugin.')(f)

    return f


def get_defs_path():
    """ Get path to HotSOS defs. """

    # source
    defs = os.path.join(get_hotsos_root(), 'defs')
    if not os.path.isdir(defs):
        # pypi
        with resources.path('hotsos', 'defs') as path:  # pylint: disable=W4902
            defs = path

    if not os.path.isdir(defs):
        # snap
        root = os.environ.get('SNAP', '/')
        defs = os.path.join(root, 'etc/hotsos/defs')

    if not os.path.exists(defs):
        raise FileNotFoundError(f"defs path {defs} not found")

    return defs


def get_templates_path():
    """ Get patch to HotSOS templates. """

    # source
    templates = os.path.join(get_hotsos_root(), 'templates')
    if not os.path.isdir(templates):
        # pypi
        # pylint: disable-next=W4902
        with resources.path('hotsos', 'templates') as path:
            templates = path

    if not os.path.isdir(templates):
        # snap
        root = os.environ.get('SNAP', '/')
        templates = os.path.join(root, 'etc/hotsos/templates')

    if not os.path.exists(templates):
        raise FileNotFoundError(f"templates path {templates} not found")

    return templates


def spinner(msg, done):
    """ CLI Spinner. """

    cli_spinner = Spinner(msg)
    while not done.is_set():
        cli_spinner.next()
        done.wait(timeout=0.1)
    cli_spinner.finish()


@contextlib.contextmanager
def progress_spinner(show_spinner, path):
    """ Launch progress spinner. """
    if not show_spinner:
        yield
        return

    sys.stdout.write(f'INFO: analysing {path}\n')
    spinner_msg = 'INFO: analysing '

    done = threading.Event()
    thread = threading.Thread(target=spinner, args=(spinner_msg, done))
    thread.start()
    try:
        yield thread
    finally:
        done.set()
        thread.join()


@dataclass(frozen=True)
class CLIArgs:  # pylint: disable=too-many-instance-attributes
    """Command line arguments."""

    data_root: click.Path
    version: bool
    defs_path: str
    templates_path: str
    all_logs: bool
    debug: bool
    quiet: bool
    save: bool
    output_format: str
    html_escape: bool
    short: bool
    very_short: bool
    force: bool
    event_tally_granularity: str
    max_logrotate_depth: int
    max_parallel_tasks: int
    list_plugins: bool
    machine_readable: bool
    output_path: str
    command_timeout: int
    sos_unpack_dir: str
    scenario: str
    event: str

    @classmethod
    def filter_kwargs(cls, **kwargs):
        """Filter the given kwargs by the names present on this
        class and produce two distinct dict's where the first
        dictionary is the fields present, the second is the
        remainder part. Used for filtering the kwargs to separate
        the arguments needed for this class and the other excess
        arguments."""

        def_fields = {field.name for field in fields(cls)}
        cleaned_kwargs = {
            key: value for key, value in kwargs.items() if key in def_fields
        }
        remainder = {k: kwargs[k] for k in set(kwargs) - set(cleaned_kwargs)}
        return (cleaned_kwargs, remainder)


def run_client(arguments, plugins_to_run, data_root_name, logmanager):
    """ Run the hotsos client inside a progress spinner. """
    with progress_spinner(
            not arguments.quiet and not arguments.debug, data_root_name):
        client = HotSOSClient(plugins_to_run)
        try:
            client.run()
        except Exception as exc:
            log.exception("An exception occurred while running "
                          "plugins")
            print('\nException running plugins:', exc)
            print('See temp log file for possible details:',
                  logmanager.temp_log_path)
            logmanager.delete_temp_file = False
            raise

        return client.summary


# pylint: disable=R0915
def main():
    @click.command(name='hotsos')
    @click.option('--event', default='',
                  help=('Filter a particular event name. Useful for '
                        'testing/debugging. Format is '
                        '<plugin name>.<group>.<event>'))
    @click.option('--scenario', default='',
                  help=('Filter a particular scenario name. Useful for '
                        'testing/debugging. Format is '
                        '<plugin name>.<group>.<scenario>'))
    @click.option('--sos-unpack-dir', default=None,
                  help=('Location used to unpack sosreports. Useful if you '
                        'want to cache the unpacked sosreport for subsequent '
                        'use. The provided path must exist.'))
    @click.option('--command-timeout', default=HotSOSConfig.command_timeout,
                  help=('Amount of time command execution will wait before '
                        'timing out and moving on.'))
    @click.option('--output-path', default=None,
                  help=('Optional path to use for saving output (with '
                        '--save).'))
    @click.option('--machine-readable', default=False, is_flag=True,
                  help="Don't format output for humans.")
    @click.option('--list-plugins', default=False, is_flag=True,
                  help='Show available plugins.')
    @click.option('--max-parallel-tasks', default=8,
                  help=('The search module will execute searches across '
                        'files in parallel. By default the number of cores '
                        'used is limited to a max of 8. You can '
                        'override that value with this option.'))
    @click.option('--max-logrotate-depth', default=7,
                  help=('Searching all available logrotate history for a '
                        'given log file can be costly so we cap the history '
                        'to this value. Only applies when --all-logs is '
                        'provided.'))
    @click.option('--event-tally-granularity', default='date',
                  help=('By default event tallies will be grouped by date, '
                        'for example when tallying occurrences of an event '
                        'in a file and displaying them in the summary. If '
                        'finer granularity is required this can be set to '
                        '"time".'))
    @click.option('--force', default=False, is_flag=True,
                  help=('By default plugins will only run if their '
                        'pre-requisites are met (i.e. they are "runnable").'
                        'It might sometimes be the case that not all '
                        'pre-requisites are available but we still want to '
                        'try running plugins. An example would be where an '
                        "application is not installed but we have it's logs. "
                        'Using this option can obviously produce '
                        'unexpected results so should be used with caution.'))
    @click.option('--very-short', default=False, is_flag=True,
                  help=('Minimal version of --short where only issue types or '
                        'bug ids are displayed with count of each (issues '
                        'only).'))
    @click.option('--short', default=False, is_flag=True,
                  help=('Filters the full summary so that it only includes '
                        'plugin known-bugs and potential-issues sections.'))
    @click.option('--html-escape', default=False, is_flag=True,
                  help=('Apply html escaping to the output so that it is safe '
                        'to display in html.'))
    @click.option('--format', '--output-format', 'output_format',
                  type=click.Choice(SUPPORTED_SUMMARY_FORMATS),
                  default='yaml',
                  show_default=True,
                  help='Summary output format.')
    @click.option('--save', '-s', default=False, is_flag=True,
                  help='Save output to a file.')
    @click.option('--quiet', default=False, is_flag=True,
                  help=('Suppress normal stderr output, only errors will be '
                        'printed.'))
    @click.option('--debug', default=False, is_flag=True,
                  help=('Provide some debug output. Logs will be printed to '
                        'stderr.'))
    @click.option('--all-logs', default=False, is_flag=True,
                  help=('Some plugins may choose to only analyse the most '
                        'recent version of a log file by default since '
                        'parsing the full history could take a lot '
                        'longer. This tells plugins that we wish to analyse '
                        'all available log history (see --max-logrotate-depth '
                        'for limits).'))
    @click.option('--templates-path', default=get_templates_path(),
                  help='Path to Jinja templates.')
    @click.option('--defs-path', default=get_defs_path(),
                  help='Path to yaml definitions (ydefs).')
    @click.option('--version', '-v', default=False, is_flag=True,
                  help='Show the version.')
    @set_plugin_options
    @click.argument('data_root', required=False, type=click.Path(exists=True))
    def cli(**kwargs):
        """
        Run this tool on a host or against a sosreport to perform
        analysis of specific applications and the host itself. A summary of
        information is generated along with any issues or known bugs detected.
        Applications are defined as plugins and support currently includes
        Openstack, Kubernetes, Ceph and more (see --list-plugins). The
        standard output format is yaml to allow easy visual inspection and
        post-processing by other tools. Other formats are also supported.

        There a three main components to this tool; the core python library,
        plugin extensions and a library of checks written in a high level
        yaml-based language.

        By default the output is printed to standard output. If saving to disk
        (--save) a directory called "hotsos-output" is created beneath which
        output is saved in all available formats e.g. yaml, json, short

        \b
        DATA_ROOT
            Path to an sosreport. If none provided, will run against local
            host.
        """  # noqa

        # Filter the kwargs to separate what should go into the CLIArgs
        # class and what not.
        (cli_kwargs, excess_kwargs) = CLIArgs.filter_kwargs(**kwargs)
        arguments = CLIArgs(**cli_kwargs)
        # Each plugin has a bool cli opt that defaults to False which if set to
        # True indicates that plugin should be run. One or more plugin can be
        # be set to True.
        plugins_to_run = {k for k, v in excess_kwargs.items()
                          if v is True}
        if plugins_to_run:
            # ensure these are always run
            plugins_to_run.add('hotsos')
            plugins_to_run.add('system')

        _version = get_version()
        if arguments.version:
            print(_version)
            return

        config = {'repo_info': get_repo_info(),
                  'force_mode': arguments.force,
                  'hotsos_version': _version,
                  'command_timeout': arguments.command_timeout,
                  'use_all_logs': arguments.all_logs,
                  'plugin_yaml_defs': arguments.defs_path,
                  'templates_path': arguments.templates_path,
                  'event_tally_granularity': arguments.event_tally_granularity,
                  'max_logrotate_depth': arguments.max_logrotate_depth,
                  'max_parallel_tasks': arguments.max_parallel_tasks,
                  'machine_readable': arguments.machine_readable,
                  'debug_mode': arguments.debug,
                  'scenario_filter': arguments.scenario,
                  'event_filter': arguments.event}
        HotSOSConfig.set(**config)

        with LoggingManager() as logmanager:
            with DataRootManager(
                arguments.data_root, sos_unpack_dir=arguments.sos_unpack_dir
            ) as drm:
                HotSOSConfig.data_root = drm.data_root
                if is_snap() and drm.data_root == '/':
                    print(SNAP_ERROR_MSG)
                    sys.exit(1)

                if arguments.debug and arguments.quiet:
                    sys.stderr.write('ERROR: cannot use both --debug and '
                                     '--quiet\n')
                    return

                # Set a name so that logs have this until real plugins are run.
                log.name = 'hotsos.cli'

                if arguments.list_plugins:
                    sys.stdout.write('\n'.join(plugintools.PLUGINS.keys()))
                    sys.stdout.write('\n')
                    return

                summary = run_client(arguments, list(plugins_to_run), drm.name,
                                     logmanager)
                if arguments.save:
                    path = summary.save(
                        drm.basename,
                        html_escape=arguments.html_escape,
                        output_path=arguments.output_path,
                    )
                    sys.stdout.write(f"INFO: output saved to {path}\n")
                else:
                    if arguments.short:
                        minimal_mode = 'short'
                    elif arguments.very_short:
                        minimal_mode = 'very-short'
                    else:
                        minimal_mode = None

                    output = summary.get_builder()
                    output.minimal(minimal_mode)
                    out = output.to(
                        fmt=arguments.output_format,
                        html_escape=arguments.html_escape
                    )

                    if out:
                        sys.stdout.write(f"{out}\n")

    cli(prog_name='hotsos')  # pylint: disable=no-value-for-parameter


def exit_if_os_version_not_supported_in_snap():
    """ Check that snap runtime requirements are met. """
    if is_snap():
        if get_os_id() != 'ubuntu':
            print("This snap has only been verified to run on Ubuntu")
            sys.exit(1)
        if get_os_version() < 20.04:
            print("This snap has only been verified to run on Focal and above")
            sys.exit(2)


if __name__ == '__main__':
    exit_if_os_version_not_supported_in_snap()
    main()



from dataclasses import dataclass, field

from hotsos.core.plugintools import PluginPartBase
from hotsos.core.host_helpers import (
    APTPackageHelper,
    InstallInfoBase,
    SystemdHelper,
)

CORE_APT = ['landscape']
SERVICE_EXPRS = [s + '[A-Za-z0-9-]*' for s in CORE_APT]


@dataclass
class LandscapeInstallInfo(InstallInfoBase):
    """ Landscape installation information. """
    apt: APTPackageHelper = field(default_factory=lambda:
                                  APTPackageHelper(core_pkgs=CORE_APT))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(service_exprs=SERVICE_EXPRS))


class LandscapeChecks(PluginPartBase):
    """ Base class for all Landscape checks. """
    plugin_name = 'landscape'
    plugin_root_index = 14

    def __init__(self, *args, **kwargs):
        super().__init__()
        LandscapeInstallInfo().mixin(self)

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        if LandscapeInstallInfo().apt.core:
            return True

        return False



from .common import InstallInfoBase
from .cli.cli import (
    CLIExecError,
    CLIHelper,
    CLIHelperFile,
)
from .config import (
    ConfigBase,
    IniConfigBase,
)
from .network import (
    NetworkPort,
    HostNetworkingHelper,
)
from .packaging import (
    DPKGVersion,
    APTPackageHelper,
    DockerImageHelper,
    SnapPackageHelper,
)
from .pebble import (  # pylint: disable=cyclic-import
    PebbleHelper,
)
from .ssl import (
    SSLCertificate,
    SSLCertificatesHelper,
)
from .systemd import (  # pylint: disable=cyclic-import
    SystemdHelper,
)
from .uptime import (  # pylint: disable=cyclic-import
    UptimeHelper,
)
from .sysctl import (
    SYSCtlFactory,
    SYSCtlConfHelper,
)
from .apparmor import (
    AAProfileFactory,
    ApparmorHelper,
)

__all__ = [
    CLIExecError.__name__,
    CLIHelper.__name__,
    CLIHelperFile.__name__,
    ConfigBase.__name__,
    IniConfigBase.__name__,
    InstallInfoBase.__name__,
    NetworkPort.__name__,
    HostNetworkingHelper.__name__,
    DPKGVersion.__name__,
    APTPackageHelper.__name__,
    DockerImageHelper.__name__,
    SnapPackageHelper.__name__,
    PebbleHelper.__name__,
    SSLCertificate.__name__,
    SSLCertificatesHelper.__name__,
    SystemdHelper.__name__,
    UptimeHelper.__name__,
    SYSCtlFactory.__name__,
    SYSCtlConfHelper.__name__,
    AAProfileFactory.__name__,
    ApparmorHelper.__name__,
]



import abc
import re
from dataclasses import dataclass, field

from hotsos.core import plugintools
from hotsos.core.host_helpers import (
    APTPackageHelper,
    InstallInfoBase,
    PebbleHelper,
    SnapPackageHelper,
    SystemdHelper,
)
from hotsos.core.plugins.openstack.openstack import OST_SUNBEAM_SNAP_NAMES
from hotsos.core.ycheck.events import EventCallbackBase, EventHandlerBase
from hotsos.core.ycheck.common import GlobalSearcherAutoRegisterBase
from hotsos.core.utils import PathFinderBase, sorted_dict

OVS_SERVICES_EXPRS = [r'ovsdb[a-zA-Z-]*',
                      r'ovs-vswitch[a-zA-Z-]*',
                      r'ovn[a-zA-Z-]*',
                      'openvswitch-switch',
                      ]
PY_CLIENT_PREFIX = r"python3?-{}\S*"
_OVS_PKGS_CORE = ['openvswitch-switch',
                  'ovn',
                  ]
OVS_PKGS_CORE = _OVS_PKGS_CORE + \
                [PY_CLIENT_PREFIX.format(p) for p in _OVS_PKGS_CORE]
_OVS_PKGS_DEPS = ['libc-bin',
                  'openvswitch',
                  'ovsdbapp',
                  'openssl',
                  'openvswitch-switch-dpdk',
                  ]
OVS_PKGS_DEPS = _OVS_PKGS_DEPS + \
                [PY_CLIENT_PREFIX.format(p) for p in _OVS_PKGS_DEPS]

OVS_SNAPS_CORE = ['microovn'] + OST_SUNBEAM_SNAP_NAMES


class PathFinder(PathFinderBase):
    """ Supports both snap and deb based installed of ovs/ovn.

    Ensure this list contains every possible path available for openvswitch/ovn
    log files.
    """
    @property
    def paths(self):
        return ['var/log',
                'var/snap/openstack-hypervisor/common/log',
                'var/snap/microovn/common/logs']


class OpenvSwitchGlobalSearchBase(GlobalSearcherAutoRegisterBase):
    """ Base class for global searcher registration for OpenvSwitch. """
    plugin_name = "openvswitch"

    @classmethod
    @abc.abstractmethod
    def paths(cls):
        """ Returns a list of one or more paths to search. """


@dataclass
class OpenvSwitchInstallInfo(InstallInfoBase):
    """ OpenvSwitch installation information. """
    apt: APTPackageHelper = field(default_factory=lambda:
                                  APTPackageHelper(core_pkgs=OVS_PKGS_CORE,
                                                   other_pkgs=OVS_PKGS_DEPS))
    pebble: PebbleHelper = field(default_factory=lambda:
                                 PebbleHelper(
                                            service_exprs=OVS_SERVICES_EXPRS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(
                                            service_exprs=OVS_SERVICES_EXPRS))
    snaps: SnapPackageHelper = field(default_factory=lambda:
                                     SnapPackageHelper(
                                            core_snaps=OVS_SNAPS_CORE))


class OpenvSwitchChecks(plugintools.PluginPartBase):
    """ OpenvSwitch checks. """
    plugin_name = "openvswitch"
    plugin_root_index = 6

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        OpenvSwitchInstallInfo().mixin(self)

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        # require at least one core package to be installed to run this plugin.
        if len(OpenvSwitchInstallInfo().apt.core) > 0:
            return True

        # This is to account for OpenStack Sunbeam
        return len(OpenvSwitchInstallInfo().snaps.core) > 0


class OpenvSwitchEventCallbackBase(EventCallbackBase):
    """ Base class for OpenvSwitch events callbacks. """
    @classmethod
    def global_event_tally_time_granularity_override(cls):
        return True

    @abc.abstractmethod
    def __call__(self):
        """ Callback method. """


class OpenvSwitchEventHandlerBase(OpenvSwitchChecks, EventHandlerBase):
    """ Base class for OpenvSwitch event handlers. """
    @property
    def summary(self):
        # mainline all results into summary root
        ret = self.run()
        if ret:
            return sorted_dict(ret)

        return None


class OVSDBTableBase():
    """
    Provides an interface to an OVSDB table. Records can be extracted from
    either 'get' or 'list' command outputs. We try 'get' first and of not found
    we search in output of 'list'.
    """
    def __init__(self, name):
        self.name = name

    @staticmethod
    def _convert_record_to_dict(record):
        """ Convert the ovsdb record dict format to a python dictionary. """
        out = {}
        if not record or record == '{}':
            return out

        expr = r'(\S+="[^"]+"|\S+=\S+),? ?'
        for _field in re.compile(expr).findall(record):
            for char in [',', '}', '{']:
                _field = _field.strip(char)

            key, _, val = _field.partition('=')
            out[key] = val.strip('"')

        return out

    @staticmethod
    def _get_cmd(table):
        raise NotImplementedError

    @staticmethod
    def _list_cmd(table):
        raise NotImplementedError

    def _fallback_query(self, column):
        """ Find first occurrence of column and return it. """
        for cmd in [self._list_cmd(self.name),
                    self._list_cmd(self.name.lower())]:
            for line in cmd():
                if not line.startswith(f'{column} '):
                    continue

                return line.partition(':')[2].strip()

        return None

    def get(self, record, column):
        """
        Try to get column using get command and failing that, try getting it
        from list.
        """
        value = self._get_cmd(self.name)(record=record, column=column)
        if not value:
            value = self._fallback_query(column=column)

        if not (value and value.startswith('{')):
            return value

        return self._convert_record_to_dict(value)

    def __getattr__(self, column):
        """ Get column for special records i.e. with key '.' """
        return self.get(record='.', column=column)



from functools import cached_property

from propertree.propertree2 import (
    PTreeLogicalGrouping,
    PTreeOverrideLiteralType,
)
from hotsos.core.exceptions import ScenarioException
from hotsos.core.issues import IssueContext
from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.common import (
    YPropertyOverrideBase,
    PropertyCacheRefResolver,
    YPropertyMappedOverrideBase,
    YDefsSection,
    YDefsContext,
)
from hotsos.core.exceptions import NotYetInitializedError


class YPropertyPriority(YPropertyOverrideBase):
    """
    Priority property. Provides support for defining a conclusion priority.

    By default all conclusions will have the same priority. This allows
    conclusions to be given different priorities. One or more conclusion can
    also be given the same priority.
    """
    override_keys = ['priority']

    def __int__(self):
        return int(self.content or 1)


class YPropertyRaises(YPropertyOverrideBase):
    """
    Raises property. Provides support for defining the issue we want to raise
    if the conclusion is matches.
    """
    override_keys = ['raises']

    @property
    def cve_id(self):
        """ optional setting. """
        return self.content.get('cve-id')

    @property
    def bug_id(self):
        """ optional setting. do this to allow querying. """
        return self.content.get('bug-id')

    @property
    def message(self):
        """ optional setting. do this to allow querying. """
        return self.content.get('message')

    def message_formatted(self, checks=None):
        """
        If a format-dict is provided this will resolve any cache references
        then format the message. Returns formatted message.

        @params checks: optional dict of YPropertyChecks objects.
        """
        fdict = self.format_dict
        if not fdict:
            return self.message

        fdict = dict(fdict)

        for key, value in fdict.items():
            if PropertyCacheRefResolver.is_valid_cache_ref(value):
                rvalue = PropertyCacheRefResolver(value,
                                                  pcrr_vars=self.context.vars,
                                                  checks=checks).resolve()
                log.debug("updating format-dict key=%s with cached %s (%s)",
                          key, value, rvalue)
                fdict[key] = rvalue

        message = self.message
        if message is not None:
            message = str(message).format(**fdict)

        return message

    @cached_property
    def format_dict(self):
        """
        Optional dict of key/val pairs used to format the message string.

        Keys that start with @ are used as references to yaml properties
        allowing us to extract cached values. Alternatively an import path
        can be specified in which case the value is imported immediately and
        optional rendering function applied.
        """
        _format_dict = self.content.get('format-dict')
        if not _format_dict:
            return {}

        fdict = {}
        for k, v in _format_dict.items():
            if PropertyCacheRefResolver.is_valid_cache_ref(v):
                # save for later parsing/extraction
                fdict[k] = v
                continue

            # process now since there is no cache to resolve
            path, _, func = v.partition(':')
            value = self.get_import(self.context, path)
            if func:
                value = PropertyCacheRefResolver.apply_renderer(value, func)

            fdict[k] = value

        return fdict

    @cached_property
    def type(self):
        """ Name of core.issues.IssueTypeBase object and will be used to raise
        an issue or bug using message as argument. """
        _type = f"hotsos.core.issues.{self.content['type']}"
        return self.get_cls(self.context, _type)


class DecisionBase():
    """ base class for decision property. """

    def get_check_item(self, name):
        checks = self.context.checks  # pylint: disable=E1101
        try:
            log.debug("%s: get_check_item() %s", self.__class__.__name__,
                      name)
            return checks[name]
        except KeyError:
            log.exception("check '%s' not found in %s", name, checks)

        return None


class DecisionLogicalGrouping(DecisionBase, PTreeLogicalGrouping):
    """ Logical grouping definition for the decision member property """
    override_autoregister = False

    def get_items(self):
        items = []
        for item in super().get_items():
            if isinstance(item, PTreeOverrideLiteralType):
                item = self.get_check_item(str(item))

            items.append(item)

        log.debug("%s: items: %s", self.__class__.__name__, items)
        return items


class YPropertyDecision(DecisionBase, YPropertyMappedOverrideBase):
    """ Decision property member of conclusions property. """
    override_keys = ['decision']
    # no members, we are using a mapping to get a custom PTreeLogicalGrouping
    override_members = []
    override_logical_grouping_type = DecisionLogicalGrouping

    @property
    def result(self):
        # pylint: disable=duplicate-code
        results = []
        try:
            stop_executon = False
            for member in self.members:
                for item in member:
                    if isinstance(item, PTreeOverrideLiteralType):
                        item = self.get_check_item(str(item))

                    result = item.result
                    results.append(result)
                    if DecisionLogicalGrouping.is_exit_condition_met('and',
                                                                     result):
                        stop_executon = True
                        break

                if stop_executon:
                    break

            result = all(results)
            log.debug("decision: %s result=%s", results, result)
            return result
        except Exception:
            log.exception("something went wrong when executing decision")
            raise


class YPropertyConclusion(YPropertyMappedOverrideBase):
    """
    Conclusion property. Provides support for defining a conclusion as a
    decision based on the outcome of one or more checks. Also supports
    defining the issue type that is raised if the conclusion if matched.

    This property is a mapping and has multiple members that form the
    properties of the conclusion.
    """
    override_keys = ['conclusion']
    override_members = [YPropertyPriority, YPropertyDecision, YPropertyRaises]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.issue = None
        # Use this to add any context to the issue. This context
        # will be retrievable as machine readable output.
        self.issue_context = IssueContext()

    @property
    def name(self):
        if hasattr(self, 'conclusion_name'):
            return getattr(self, 'conclusion_name')

        return None

    def reached(self, checks):
        """
        Return True/False result of this conclusion and prepare issue info.
        """
        log.debug("running conclusion %s", self.name)
        log.debug("decision:start")
        result = self.decision.result
        log.debug("decision:end")
        if not result:
            return False

        cve_id = self.raises.cve_id
        bug_id = self.raises.bug_id
        bug_type = self.raises.type.ISSUE_TYPE
        is_bug_type = bug_type in ('bug', 'cve')
        if cve_id or bug_id or is_bug_type:
            if not (all([bug_id, bug_type == 'bug']) or
                    all([cve_id, bug_type == 'cve'])):
                msg = (f"both cve-id/bug-id (current={bug_id or cve_id}) and "
                       f"bug type (current={bug_type}) required in order to "
                       "raise a bug")
                raise ScenarioException(msg)

        message = self.raises.message_formatted(checks=checks)
        if is_bug_type:
            self.issue = self.raises.type(bug_id or cve_id, message)
        else:
            self.issue = self.raises.type(message)

        return result


class YPropertyConclusions(YPropertyOverrideBase):
    """ Conclusions property.

    The property contains one or more conclusion which are loaded by iterating
    over the object.
    """
    override_keys = ['conclusions']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._initialised = False
        self.conclusion_context = YDefsContext()

    def initialise(self, ypvars, checks):
        """
        Perform initialisation tasks for this set of conclusions.
        """
        self.conclusion_context.update({'vars': ypvars, 'checks': checks})
        self._initialised = True

    @cached_property
    def _conclusions(self):
        log.debug("parsing conclusions section")
        if not self._initialised:
            raise NotYetInitializedError("conclusions not yet initialised")

        resolved = []
        for name, content in self.content.items():
            s = YDefsSection(self.override_name,
                             {name: {'conclusion': content}},
                             override_handlers=self.root.override_handlers,
                             resolve_path=self.override_path,
                             context=self.conclusion_context)
            for c in s.leaf_sections:
                c.conclusion.conclusion_name = c.name
                resolved.append(c.conclusion)

        return resolved

    def __iter__(self):
        log.debug("iterating over conclusions")
        for c in self._conclusions:
            log.debug("returning conclusion %s", c.name)
            yield c



from hotsos.core import plugintools


class StorageBase(plugintools.PluginPartBase):
    """ Base class for all storage checks. """
    plugin_name = 'storage'
    plugin_root_index = 9

    @classmethod
    def is_runnable(cls):
        raise NotImplementedError



import abc
import os
import re
import sys
from datetime import datetime
from dataclasses import dataclass, field
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.factory import FactoryBase
from hotsos.core.host_helpers import (
    APTPackageHelper,
    CLIHelper,
    CLIHelperFile,
    DPKGVersion,
    InstallInfoBase,
    HostNetworkingHelper,
    PebbleHelper,
    SnapPackageHelper,
    SystemdHelper,
    IniConfigBase,
)
from hotsos.core.log import log
from hotsos.core.plugins.kernel.net import Lsof
from hotsos.core.plugins.storage import StorageBase
from hotsos.core.plugins.storage.bcache import BcacheBase
from hotsos.core.plugins.storage.ceph.daemon import CephOSD
from hotsos.core.plugins.storage.ceph.cluster import CephCluster
from hotsos.core.search import (
    FileSearcher,
    SequenceSearchDef,
    SearchDef
)
from hotsos.core.ycheck.events import EventCallbackBase
from hotsos.core.utils import PathFinderBase

CEPH_SERVICES_EXPRS = [r"ceph-[a-z0-9-]+",
                       r"rados[a-z0-9-:]+",
                       r"rbd[a-z0-9-:]+",
                       r"microceph.(mon|mgr|mds|osd|rgw|rbd[a-z-]*)"]
CEPH_PKGS_CORE = [r"ceph",
                  r"rados",
                  r"rbd",
                  ]
CEPH_PKGS_OTHER = []
# Add in clients/deps
for ceph_pkg in CEPH_PKGS_CORE:
    CEPH_PKGS_OTHER.append(rf"python3?-{ceph_pkg}\S*")

CEPH_SNAPS_CORE = [r'microceph']

# NOTE(tpsilva): when updating this list, refer to the supported Ceph
# versions for Ubuntu page:
# https://ubuntu.com/ceph/docs/supported-ceph-versions
CEPH_EOL_INFO = {
    'squid': datetime(2036, 4, 30),
    'reef': datetime(2025, 4, 30),
    'quincy': datetime(2032, 4, 30),
    'pacific': datetime(2024, 4, 30),
    'octopus': datetime(2030, 4, 30),
    'nautilus': datetime(2021, 2, 28),
    'mimic': datetime(2022, 4, 30),
    'luminous': datetime(2028, 4, 30),
    'jewel': datetime(2024, 4, 30)
}

CEPH_REL_INFO = {
    'ceph-common': {
        'squid': '19.0',
        'reef': '18.0',
        'quincy': '17.0',
        'pacific': '16.0',
        'octopus': '15.0',
        'nautilus': '14.0',
        'mimic': '13.0',
        'luminous': '12.0',
        'kraken': '11.0',
        'jewel': '10.0'},
}


class PathFinder(PathFinderBase):
    """ Supports both snap and deb based installed of ceph.

    Ensure this list contains every possible path available for ceph
    log files.
    """
    @property
    def paths(self):
        return ['var/log/ceph',
                'var/snap/microceph/common/logs']


def csv_to_set(f):
    """
    Decorator used to convert a csv string to a set().
    """
    def csv_to_set_inner(*args, **kwargs):
        val = f(*args, **kwargs)
        if val is not None:
            return set(v.strip(',') for v in val.split())

        return set([])

    return csv_to_set_inner


class CephConfig(IniConfigBase):
    """
    Ceph config.

    Adds support for specific format used in Ceph config files.
    """
    def __init__(self, *args, **kwargs):
        path = os.path.join(HotSOSConfig.data_root, 'etc/ceph/ceph.conf')
        if not os.path.exists(path):
            path = os.path.join(HotSOSConfig.data_root,
                                'var/snap/microceph/current/conf/ceph.conf')

        super().__init__(*args, path=path, **kwargs)

    def get(self, key, *args, **kwargs):
        """
        First to get value for key and if not found, try alternative key format
        i.e. ceph config like 'ceph osd debug' and 'ceph_osd_debug' are both
        valid and equivalent so we try both by converting the key name. If that
        still does not match we look for key in this object instance since it
        may have been provided as a property.
        """
        val = super().get(key, *args, **kwargs)
        if val is not None:
            return val

        orig_key = key
        if ' ' in key:
            key = key.replace(' ', '_')
        else:
            key = key.replace('_', ' ')

        val = super().get(key, *args, **kwargs)
        if val is not None:
            return val

        if hasattr(self, orig_key):
            return getattr(self, orig_key)

        return None

    @property
    @csv_to_set
    def cluster_network_set(self):
        return self.get('cluster network')

    @property
    @csv_to_set
    def public_network_set(self):
        return self.get('public network')


@dataclass
class CephInstallInfo(InstallInfoBase):
    """ Ceph installation information. """
    apt: APTPackageHelper = field(default_factory=lambda:
                                  APTPackageHelper(
                                                core_pkgs=CEPH_PKGS_CORE,
                                                other_pkgs=CEPH_PKGS_OTHER))
    pebble: PebbleHelper = field(default_factory=lambda:
                                 PebbleHelper(
                                            service_exprs=CEPH_SERVICES_EXPRS))
    snaps: SnapPackageHelper = field(default_factory=lambda:
                                     SnapPackageHelper(
                                                core_snaps=CEPH_SNAPS_CORE))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(
                                            service_exprs=CEPH_SERVICES_EXPRS))


class CephChecks(StorageBase):
    """ Ceph Checks. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ceph_config = CephConfig()
        self.bcache = BcacheBase()
        self.cluster = CephCluster()
        CephInstallInfo().mixin(self)

    @property
    def summary_subkey_include_default_entries(self):
        return True

    @property
    def summary_subkey(self):
        return 'ceph'

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        ceph = CephInstallInfo()
        if ceph.apt.core or ceph.snaps.core:
            return True

        return False

    @cached_property
    def release_name(self):
        relname = 'unknown'

        pkg = 'ceph-common'
        pkg_version = None
        if self.apt.core and pkg in self.apt.core:
            pkg_version = self.apt.core[pkg]
        elif self.snaps.core and 'microceph' in self.snaps.core:
            pkg_version = self.snaps.core['microceph']['version']
            pkg_version = pkg_version.partition("+snap")[0]

        if pkg_version is not None:
            for rel, ver in sorted(CEPH_REL_INFO[pkg].items(),
                                   key=lambda i: i[1], reverse=True):
                if pkg_version > DPKGVersion(ver):
                    relname = rel
                    break

        return relname

    @cached_property
    def days_to_eol(self):
        if self.release_name == 'unknown':
            return None

        eol = CEPH_EOL_INFO[self.release_name]
        today = datetime.utcfromtimestamp(int(CLIHelper().date()))
        delta = (eol - today).days
        return delta

    def _get_bind_interfaces(self, iface_type):
        """
        For the given config network type determine what interface ceph is
        binding to.

        @param iface_type: cluster or public
        """
        net = self.ceph_config.get(f'{iface_type} network')
        addr = self.ceph_config.get(f'{iface_type} addr')
        if not any([net, addr]):
            return {}

        nethelp = HostNetworkingHelper()
        port = None
        if net:
            port = nethelp.get_interface_with_addr(net)
        elif addr:
            port = nethelp.get_interface_with_addr(addr)

        if port:
            return {iface_type: port}

        return None

    @cached_property
    def _ceph_bind_interfaces(self):
        """
        Returns a dict of network interfaces used by Ceph daemons on this host.
        The dict has the form {<type>: [<port>, ...]}
        """
        interfaces = {}
        for iface_type in ['cluster', 'public']:
            ret = self._get_bind_interfaces(iface_type)
            if ret:
                interfaces.update(ret)

        return interfaces

    @property
    def bind_interfaces(self):
        return self._ceph_bind_interfaces

    @cached_property
    def bind_interface_names(self):
        """
        Returns a list of names for network interfaces used by Ceph daemons on
        this host.
        """
        names = [iface.name for iface in self.bind_interfaces.values()]
        return ', '.join(list(set(names)))

    @cached_property
    def local_osds(self):
        """
        Returns a list of CephOSD objects for osds found on the local host.
        """
        osds = []

        s = FileSearcher()
        sd = SequenceSearchDef(start=SearchDef(r"^=+\s+osd\.(\d+)\s+=+.*"),
                               body=SearchDef([r"\s+osd\s+(fsid)\s+(\S+)\s*",
                                               r"\s+(devices)\s+([\S]+)\s*"]),
                               tag="ceph-lvm")
        with CLIHelperFile() as cli:
            fout = cli.ceph_volume_lvm_list()
            s.add(sd, path=fout)
            for results in s.run().find_sequence_sections(sd).values():
                osdid = None
                fsid = None
                dev = None
                for result in results:
                    if result.tag == sd.start_tag:
                        osdid = int(result.get(1))
                    elif result.tag == sd.body_tag:
                        if result.get(1) == "fsid":
                            fsid = result.get(2)
                        elif result.get(1) == "devices":
                            dev = result.get(2)

                osds.append(CephOSD(osdid, fsid, dev))

        return osds

    @cached_property
    def local_osds_use_bcache(self):
        """
        Returns True if any local osds are using bcache devices.
        """
        for osd in self.local_osds:
            if self.bcache.is_bcache_device(osd.device):
                return True

        return False

    @cached_property
    def local_osds_devtypes(self):
        return [osd.devtype for osd in self.local_osds]

    @cached_property
    def bluestore_enabled(self):
        """
        If any of the following are enabled in ceph.conf (by the charm) it
        indicates that bluestore=True.
        """
        bs_key_vals = {('enable experimental unrecoverable data corrupting '
                        'features'): 'bluestore',
                       'osd objectstore': 'bluestore'}
        bs_keys = ['bluestore block wal size', 'bluestore block db size',
                   r'bluestore compression .+']

        for keys in self.ceph_config.all_keys:
            for conf_key in keys:
                if conf_key in bs_keys:
                    return True

                for key in bs_keys:
                    if re.compile(key).match(conf_key):
                        return True

        for key, val in bs_key_vals.items():
            conf_val = self.ceph_config.get(key)
            if conf_val and val in conf_val:
                return True

        return False

    @cached_property
    def has_interface_errors(self):
        """
        Checks if any network interfaces used by Ceph are showing packet
        errors.

        Returns True if errors found otherwise False.
        """
        for port in self.bind_interfaces.values():
            for stats in port.stats.values():
                if stats.get('errors'):
                    return True

        return False

    @cached_property
    def linked_with_tcmalloc(self):
        """
        Checks that each ceph-osd process has libtcmalloc linked.

        Returns True if every OSD has it linked, otherwise False.
        """
        osds = {}
        tcmalloc_osds = 0
        for row in Lsof():
            if row.COMMAND == 'ceph-osd':
                osds[row.PID] = 1
                if re.search("libtcmalloc", row.NAME):
                    tcmalloc_osds += 1

        return len(osds) == tcmalloc_osds


class CephDaemonCommand():
    """
    This class is used to run a ceph daemon command that must be supported by
    CLIHelper. Attributes of the output can then be retrieved by calling them
    on the returned object.
    """
    def __init__(self, command, *args, **kwargs):
        self.command = command
        self.output = getattr(CLIHelper(), command)(*args, **kwargs)

    def __getattr__(self, name):
        if name in self.output:
            return self.output[name]

        raise AttributeError(f"{name} not found in output of {self.command}")


class CephDaemonConfigShow():
    """ Interface to ceph daemon config show command. """
    def __init__(self, osd_id):
        self.cmd = CephDaemonCommand('ceph_daemon_osd_config_show',
                                     osd_id=osd_id)

    def __getattr__(self, name):
        return getattr(self.cmd, name)


class CephDaemonDumpMemPools():
    """ Interface to ceph daemon osd dump mempools. """
    def __init__(self, osd_id):
        self.cmd = CephDaemonCommand('ceph_daemon_osd_dump_mempools',
                                     osd_id=osd_id)

    def __getattr__(self, name):
        val = getattr(self.cmd, 'mempool')
        if val:
            return val.get('by_pool', {}).get(name, {}).get('items')

        return None


class CephDaemonAllOSDsCommand():
    """
    This class is used to run CephDaemonCommand for all local OSDs.
    """
    def __init__(self, command):
        self.checks_base = CephChecks()
        self.command = command

    def __getattr__(self, name=None):
        """
        First instantiates the requested ceph daemon command handler then
        retrieves the requested attribute/operand and returns as a list of
        unique values.
        """
        vals = set()
        for osd in self.checks_base.local_osds:
            try:
                config = getattr(sys.modules[__name__],
                                 self.command)(osd_id=osd.id)
            except ImportError:
                log.warning("no ceph daemon command handler found for '%s'")
                break

            if hasattr(config, name):
                vals.add(getattr(config, name))

        return list(vals)


class CephDaemonAllOSDsFactory(FactoryBase):
    """
    A factory interface to allow dynamic access to ceph daemon commands and
    attributes of the output.
    """

    def __getattr__(self, command):
        return CephDaemonAllOSDsCommand(command)


class CephEventCallbackBase(EventCallbackBase):
    """ Base class for ceph event callbacks. """

    @abc.abstractmethod
    def __call__(self):
        """ Callback method. """




# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' barbican/common/exception.py
BARBICAN_EXCEPTIONS = [
    "BarbicanException",
    "BarbicanHTTPException",
    "MissingArgumentError",
    "MissingMetadataField",
    "InvalidMetadataRequest",
    "InvalidMetadataKey",
    "InvalidSubjectDN",
    "InvalidContainer",
    "InvalidExtensionsData",
    "InvalidCMCData",
    "InvalidPKCS10Data",
    "InvalidCertificateRequestType",
    "CertificateExtensionsNotSupported",
    "FullCMCNotSupported",
    "NotFound",
    "ConstraintCheck",
    "NotSupported",
    "Invalid",
    "NoDataToProcess",
    "LimitExceeded",
    "InvalidObject",
    "PayloadDecodingError",
    "UnsupportedField",
    "FeatureNotImplemented",
    "StoredKeyContainerNotFound",
    "StoredKeyPrivateKeyNotFound",
    "ProvidedTransportKeyNotFound",
    "InvalidCAID",
    "CANotDefinedForProject",
    "QuotaReached",
    "InvalidParentCA",
    "SubCAsNotSupported",
    "SubCANotCreated",
    "CannotDeleteBaseCA",
    "UnauthorizedSubCA",
    "CannotDeletePreferredCA",
    "BadSubCACreationRequest",
    "SubCACreationErrors",
    "SubCADeletionErrors",
    "PKCS11Exception",
    "P11CryptoPluginKeyException",
    "P11CryptoPluginException",
    "P11CryptoKeyHandleException",
    "P11CryptoTokenException",
    "MultipleStorePreferredPluginMissing",
    "MultipleStorePluginStillInUse",
    "MultipleSecretStoreLookupFailed",
    "MultipleStoreIncorrectGlobalDefault",
    "MultipleStorePluginValueMissing",
]



import abc
from collections import UserList

from hotsos.core.log import log
from hotsos.core.host_helpers.cli.common import BinCmd, FileCmd, CmdOutput
from hotsos.core.host_helpers.exceptions import (
    CLIExecError,
    SourceNotFound,
)

OVSAliases = ['openstack-hypervisor.', 'microovn.']


class OVSVSCtlCmdsBase(UserList):
    """ Base for generating ovs-vsctl command variants. """
    CMD = None
    ARGS = None
    SINGLELINE = False

    @abc.abstractmethod
    def get_sos_path(self, prefix):
        """ Return path to sosereport file output. """

    def __init__(self):
        cmd = f'ovs-vsctl {self.CMD}'
        if self.ARGS:
            cmd += f' {self.ARGS}'

        # binary commands
        cmds = [BinCmd(cmd, singleline=self.SINGLELINE)]
        cmds.extend([BinCmd(f'{prefix}{cmd}', singleline=self.SINGLELINE)
                    for prefix in OVSAliases])
        # file-based commands
        cmds.append(FileCmd(self.get_sos_path(prefix=''),
                            singleline=self.SINGLELINE))
        cmds.extend([FileCmd(self.get_sos_path(prefix=prefix),
                             singleline=self.SINGLELINE)
                    for prefix in OVSAliases])
        super().__init__(cmds)


class OVSVSCtlGetCmds(OVSVSCtlCmdsBase):
    """ Generate ovs-vsctl get command variants. """
    CMD = 'get'
    ARGS = '{table} {record} {column}'
    SINGLELINE = True

    def get_sos_path(self, prefix):
        return (f'sos_commands/openvswitch/{prefix}ovs-vsctl_-t_5_'
                f'{self.CMD}_{self.ARGS.replace(" ", "_")}')


class OVSVSCtlListCmds(OVSVSCtlCmdsBase):
    """ Generate ovs-vsctl list command variants. """
    CMD = 'list'
    ARGS = '{table}'

    def get_sos_path(self, prefix):
        return (f'sos_commands/openvswitch/{prefix}ovs-vsctl_-t_5_{self.CMD}_'
                f'{self.ARGS.replace(" ", "_")}')


class OVSVSCtlListBrCmds(OVSVSCtlCmdsBase):
    """ Generate ovs-vsctl list-br command variants. """
    CMD = 'list-br'

    def get_sos_path(self, prefix):
        return f'sos_commands/openvswitch/{prefix}ovs-vsctl_-t_5_{self.CMD}'


class OVSAppCtlBinCmd(BinCmd):
    """ Implements ovs-appctl binary command. """

    def __call__(self, *args, **kwargs):
        # Set defaults for optional args
        for key in ['flags', 'args']:
            if key not in kwargs:
                kwargs[key] = ''

        return super().__call__(*args, **kwargs)


class OVSAppCtlFileCmd(FileCmd):
    """ Implements ovs-appctl file-based command. """

    def __init__(self, *args, prefix=None, **kwargs):
        if not prefix:
            prefix = ''

        path = (f'sos_commands/openvswitch/{prefix}ovs-appctl_'
                '{command}{flags}{args}')
        super().__init__(path, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        for key in ['flags', 'args']:
            if key in kwargs:
                kwargs[key] = f'_{kwargs[key]}'
            else:
                kwargs[key] = ''

        if 'args' in kwargs and kwargs['command'].startswith('dpctl/'):
            # e.g. this would be for dpctl datapath
            kwargs['args'] = kwargs['args'].replace('@', '_')

        kwargs['command'] = kwargs['command'].replace('/', '.')
        try:
            out = super().__call__(*args, **kwargs)
        except SourceNotFound:
            log.debug("%s: source not found: %s", self.__class__.__name__,
                      self.path)
            raise

        return out


class OVSAppCtlCmds(UserList):
    """ Generate ovs-appctl command variants. """

    def __init__(self):
        cmd = 'ovs-appctl'
        # binary commands
        cmds = [OVSAppCtlBinCmd(cmd + ' {command} {flags} {args}')]
        cmds.extend([OVSAppCtlBinCmd(f"{prefix}{cmd}" +
                                     ' {command} {flags} {args}')
                    for prefix in OVSAliases])
        # file-based commands
        cmds.append(OVSAppCtlFileCmd())
        cmds.extend([OVSAppCtlFileCmd(prefix=prefix)
                    for prefix in OVSAliases])
        super().__init__(cmds)


OFPROTOCOL_VERSIONS = [
    "OpenFlow15",
    "OpenFlow14",
    "OpenFlow13",
    "OpenFlow12",
    "OpenFlow11",
    "OpenFlow10",
]


class OVSOFCtlBinCmd(BinCmd):
    """ Implementation of ovs-ofctl binary command. """
    BIN_CMD = 'ovs-ofctl'

    def __init__(self, *args, prefix=None, **kwargs):
        self.prefix = prefix
        super().__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        First try without specifying protocol version. If error is raised
        try with different versions until we get a result.
        """
        self.cmd = f"{self.BIN_CMD} {self.cmd}"
        if self.prefix:
            self.cmd = f'{self.prefix}{self.cmd}'

        try:
            return super().__call__(*args, **kwargs)
        except CLIExecError:
            log.debug("%s: command with no protocol version failed",
                      self.__class__.__name__)

        # If the command raised an exception it will have been caught by the
        # catch_exceptions decorator and [] returned. We have no way of knowing
        # if that was the actual return or an exception was raised so we just
        # go ahead and retry with specific OF versions until we get a result.
        for ver in OFPROTOCOL_VERSIONS:
            log.debug("%s: trying again with protocol version %s",
                      self.__class__.__name__, ver)
            self.reset()
            self.cmd = f"{self.BIN_CMD} -O {ver} {self.cmd}"
            if self.prefix:
                self.cmd = f'{self.prefix}{self.cmd}'

            try:
                return super().__call__(*args, **kwargs)
            except CLIExecError:
                log.debug("%s: command with protocol version %s failed",
                          self.__class__.__name__, ver)

        return CmdOutput([])


class OVSOFCtlFileCmd(FileCmd):
    """ Implementation of ovs-ofctl file-based command. """

    def __init__(self, *args, prefix=None, **kwargs):
        if not prefix:
            prefix = ''

        path = (f'sos_commands/openvswitch/{prefix}ovs-ofctl'
                '{ofversion}_{command}_{args}')
        super().__init__(path, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        We do this in reverse order to bin command since it won't actually
        raise an error.
        """
        for ver in OFPROTOCOL_VERSIONS:
            log.debug("%s: trying again with protocol version %s",
                      self.__class__.__name__, ver)
            self.reset()
            try:
                kwargs['ofversion'] = f'_-O_{ver}'
                out = super().__call__(*args, **kwargs)
                if out.value and 'version negotiation failed' in out.value[0]:
                    continue

                return out
            except SourceNotFound:
                pass

        try:
            kwargs['ofversion'] = ''
            return super().__call__(*args, **kwargs)
        except SourceNotFound:
            log.debug("%s: command with no protocol version failed",
                      self.__class__.__name__)

        return None


class OVSOFCtlCmds(UserList):
    """ Generate ovs-ofctl command variants. """

    def __init__(self):
        # binary
        cmds = [OVSOFCtlBinCmd('{command} {args}')]
        cmds.extend([OVSOFCtlBinCmd('{command} {args}', prefix=prefix)
                    for prefix in OVSAliases])
        # file-based
        cmds.append(OVSOFCtlFileCmd())
        cmds.extend([OVSOFCtlFileCmd(prefix=prefix)
                    for prefix in OVSAliases])
        super().__init__(cmds)



from dataclasses import dataclass, field

from hotsos.core.host_helpers import (
    APTPackageHelper,
    InstallInfoBase,
    SnapPackageHelper,
    SystemdHelper,
)
from hotsos.core.plugintools import PluginPartBase

MAAS_CORE_APT = ['maas']
MAAS_APT_DEPS = ['isc-dhcp', 'bind9', 'postgres']
MAAS_CORE_SNAPS = MAAS_CORE_APT

SERVICE_EXPRS = [s + '[A-Za-z0-9-]*' for s in MAAS_CORE_APT + MAAS_APT_DEPS]


@dataclass
class MAASInstallInfo(InstallInfoBase):
    """ MAAS installation information. """
    apt: APTPackageHelper = field(default_factory=lambda:
                                  APTPackageHelper(core_pkgs=MAAS_CORE_APT,
                                                   other_pkgs=MAAS_APT_DEPS))
    snaps: SnapPackageHelper = field(default_factory=lambda:
                                     SnapPackageHelper(
                                         core_snaps=MAAS_CORE_SNAPS))
    systemd: SystemdHelper = field(default_factory=lambda:
                                   SystemdHelper(service_exprs=SERVICE_EXPRS,
                                                 ps_allow_relative=False))


class MAASChecks(PluginPartBase):
    """ MAAS checks. """
    plugin_name = 'maas'
    plugin_root_index = 13

    def __init__(self, *args, **kwargs):
        super().__init__()
        MAASInstallInfo().mixin(self)

    @classmethod
    def is_runnable(cls):
        """
        Determine whether or not this plugin can and should be run.

        @return: True or False
        """
        maas = MAASInstallInfo()
        if maas.apt.core or maas.snaps.core:
            return True

        return False



from hotsos.core.plugins.openstack.common import (
    OpenstackEventCallbackBase,
    OpenstackEventHandlerBase,
)
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)

EXT_EVENT_META = {'network-vif-plugged': {'stages_keys':
                                          ['Preparing', 'Received',
                                           'Processing']},
                  'network-changed': {'stages_keys': ['Received',
                                                      'Refreshing']}}


class ExternalEventsCallback(OpenstackEventCallbackBase):
    """ Implements Openstack Nova external events callback.

    @return: dictionary with count of succeeded and list of failed (incomplete)
             external events grouped by event type.
    """
    event_group = 'nova.external-events'
    event_names = ['network-changed', 'network-vif-plugged']

    def __call__(self, event):
        data = {}
        for result in event.results:
            instance_id = result.get(1)
            stage = result.get(2)
            event_id = result.get(3)
            # If the same event/port is used for a new instance we reset.
            if (event_id in data and
                    data[event_id]['instance_id'] == instance_id):
                stages = data[event_id]['stages']
                stages.add(stage)
                if (not
                        stages.difference(EXT_EVENT_META[event.name]
                                          ['stages_keys'])):
                    data[event_id]['complete'] = True
            else:
                data[event_id] = {'complete': False,
                                  'stages': {stage},
                                  'instance_id': instance_id}

        out = {'succeeded': 0,
               'failed': []}
        for e, info in data.items():
            if info['complete']:
                out['succeeded'] += 1
            else:
                x = {'port': e, 'instance': info['instance_id']}
                out['failed'].append(x)

        if not out['succeeded']:
            del out['succeeded']

        if not out['failed']:
            del out['failed']

        return out


class NovaExternalEventChecks(OpenstackEventHandlerBase):
    """ Implements Openstack Nova external events handler. """
    event_group = 'nova.external-events'
    summary_part_index = 1

    @summary_entry('os-server-external-events',
                   get_min_available_entry_index() + 4)
    def summary_os_server_external_events(self):
        return self.run()



import json
import subprocess

from hotsos.core.log import log

CLI_COMMON_EXCEPTIONS = (OSError, subprocess.CalledProcessError,
                         subprocess.TimeoutExpired,
                         json.JSONDecodeError)


class CLIExecError(Exception):
    """ Exception raised when execution of a command files. """
    def __init__(self, message, return_value=None):
        """
        @param message: information message about the exception
        @param return_value: default return value that a command
                             should return if execution fails.
        """
        self.message = message
        self.return_value = return_value


def catch_exceptions(*exc_types):
    def catch_exceptions_inner1(f):
        def catch_exceptions_inner2(*args, **kwargs):
            try:
                return f(*args, **kwargs)
            except exc_types as exc:
                msg = f"{type(exc)}: {exc}"
                if isinstance(exc, subprocess.TimeoutExpired):
                    log.info(msg)
                else:
                    log.debug(msg)

                if isinstance(exc, json.JSONDecodeError):
                    raise CLIExecError(str(exc), return_value={}) from exc

                if isinstance(exc, subprocess.CalledProcessError):
                    raise CLIExecError(exc.output, return_value={}) from exc

                raise CLIExecError(str(exc), return_value=[]) from exc

        return catch_exceptions_inner2

    return catch_exceptions_inner1


class SourceNotFound(Exception):
    """ Exception raised when a file-based command is not found. """
    def __init__(self, path):
        self.path = path

    def __repr__(self):
        return f"source path '{self.path}' not found"


class CommandNotFound(Exception):
    """ Exception raised when command is not found. """
    def __init__(self, cmd, msg):
        self.msg = f"command '{cmd}' not found in catalog: '{msg}'"

    def __str__(self):
        return self.msg



from hotsos.core.plugins.sosreport import SOSReportChecks
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class SOSReportSummary(SOSReportChecks):
    """ Implementation of SOSReport summary. """
    summary_part_index = 0

    # REMINDER: Common entries are implemented in
    #           plugintools.ApplicationSummaryBase. Only customisations are
    #           implemented here. See
    #           plugintools.get_min_available_entry_index() for an explanation
    #           on how entry indices are managed.

    @summary_entry('plugin-timeouts', get_min_available_entry_index())
    def summary_plugin_timeouts(self):
        if self.timed_out_plugins:
            return self.timed_out_plugins

        return None




# sed -rn 's/^class\s+(\S+)\(.+/    "\1",/p' masakari/exception.py
MASAKARI_EXCEPTIONS = [
    "ConvertedException",
    "MasakariException",
    "APIException",
    "APITimeout",
    "Conflict",
    "Invalid",
    "InvalidName",
    "InvalidInput",
    "InvalidAPIVersionString",
    "MalformedRequestBody",
    "NotFound",
    "ConfigNotFound",
    "Forbidden",
    "AdminRequired",
    "PolicyNotAuthorized",
    "PasteAppNotFound",
    "InvalidContentType",
    "VersionNotFoundForAPIMethod",
    "InvalidGlobalAPIVersion",
    "ApiVersionsIntersect",
    "ValidationError",
    "InvalidSortKey",
    "MarkerNotFound",
    "FailoverSegmentNotFound",
    "HostNotFound",
    "NotificationNotFound",
    "FailoverSegmentNotFoundByName",
    "HostNotFoundByName",
    "ComputeNotFoundByName",
    "FailoverSegmentExists",
    "HostExists",
    "Unauthorized",
    "ObjectActionError",
    "OrphanedObjectError",
    "DuplicateNotification",
    "HostOnMaintenanceError",
    "HostRecoveryFailureException",
    "InstanceRecoveryFailureException",
    "SkipInstanceRecoveryException",
    "SkipProcessRecoveryException",
    "SkipHostRecoveryException",
    "ProcessRecoveryFailureException",
    "DBNotAllowed",
    "FailoverSegmentInUse",
    "HostInUse",
    "ReservedHostsUnavailable",
    "LockAlreadyAcquired",
    "IgnoreInstanceRecoveryException",
    "HostNotFoundUnderFailoverSegment",
    "InstanceEvacuateFailed",
    "FailoverSegmentDisabled",
]



from . import (
    summary,
    service_features,
    service_network_checks,
    vm_info,
    nova_external_events,
    sunbeam,
)
from .agent import (
    events,
    exceptions,
)

__all__ = [
    events,
    exceptions,
    nova_external_events,
    summary,
    service_features,
    service_network_checks,
    sunbeam,
    vm_info,
    ]



import json
import os
import re
import subprocess
from dataclasses import dataclass, field, fields

import yaml
from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers.exceptions import (
    catch_exceptions,
    CLI_COMMON_EXCEPTIONS,
    SourceNotFound,
)
from hotsos.core.log import log


@dataclass(frozen=True)
class CmdOutput():
    """ Representation of the output of a command. """

    # Output value.
    value: str
    # Optional command source path.
    source: str = None


def run_pre_exec_hooks(f):
    """ pre-exec hooks are run before running __call__ method.

    These hooks are not expected to return anything and are used to manipulate
    the instance variables used by the main __call__ method.
    """
    def run_pre_exec_hooks_inner(self, *args, **kwargs):
        hook = self.hooks.get("pre-exec")
        if hook:
            # no return expected
            hook(*args, **kwargs)

        return f(self, *args, **kwargs)

    return run_pre_exec_hooks_inner


def run_post_exec_hooks(f):
    """ post-exec hooks are run after running __call__ method and take its
    output as input.
    """
    def run_post_exec_hooks_inner(self, *args, **kwargs):
        out = f(self, *args, **kwargs)
        hook = self.hooks.get("post-exec")
        if hook:
            out = hook(out, *args, **kwargs)

        return out

    return run_post_exec_hooks_inner


def reset_command(f):
    """
    This should be run by all commands as their last action after all/any hooks
    have run.
    """
    def reset_command_inner(self, *args, **kwargs):
        out = f(self, *args, **kwargs)
        self.reset()
        return out

    return reset_command_inner


@dataclass
class CmdBase:
    """ Base class for all command source types.

    Provides a way to save original state and restore to that state.
    """
    def __post_init__(self):
        self.hooks = {}
        # make a copy of field original values.
        for f in fields(self):
            setattr(self, 'original_' + f.name, f.type(getattr(self, f.name)))

    def get_original_attr_value(self, name):
        return getattr(self, 'original_' + name)

    def reset(self):
        """ Reset fields to original values. """
        for f in fields(self):
            setattr(self, f.name, f.type(getattr(self, 'original_' + f.name)))

    @classmethod
    def safe_readlines(cls, path):
        with open(path, 'r', encoding='utf-8', errors="surrogateescape") as fd:
            return fd.readlines()

    def register_hook(self, name, f):
        """
        Implementations of this class can register hooks to
        be run when the __call__ method is executed. Currently
        supported hooks include:

            * pre-exec - run before __call__ method.
            * post-exec - run after __call__ method and take its output as
                          input.
        """
        self.hooks[name] = f


@dataclass
class FileCmdBase(CmdBase):
    """
    State used to execute a file-based command i.e. a command whose output is
    already saved in a file.
    """
    path: str
    json_decode: bool = False
    yaml_decode: bool = False
    singleline: bool = False
    decode_error_handling: str = None

    def __post_init__(self):
        self.path = os.path.join(HotSOSConfig.data_root, self.path)
        super().__post_init__()


@dataclass
class BinCmdBase(CmdBase):
    """ State used to execute a binary command. """
    cmd: str
    json_decode: bool = False
    yaml_decode: bool = False
    singleline: bool = False
    cmd_extras: list = field(default_factory=lambda: [])


class BinCmd(BinCmdBase):
    """ Implements binary command execution. """
    TYPE = "BIN"

    @catch_exceptions(*CLI_COMMON_EXCEPTIONS)
    @reset_command
    @run_post_exec_hooks
    @run_pre_exec_hooks
    def __call__(self, *args, skip_json_decode=False, **kwargs):
        cmd = self.cmd
        if args:
            cmd = cmd.format(*args)

        if kwargs:
            cmd = cmd.format(**kwargs)

        _cmd = cmd.split() + self.cmd_extras
        log.debug("executing cli command: %s", _cmd)
        out = subprocess.run(_cmd, timeout=HotSOSConfig.command_timeout,
                             capture_output=True, check=False)
        output = out.stdout
        if out.stderr:
            output += out.stderr

        if out.returncode != 0:
            log.info("command '%s' exited with non-zero code '%s'", cmd,
                     out.returncode)

        try:
            output = output.decode('UTF-8')
        except UnicodeDecodeError:
            log.exception("failed to decode command output for '%s'", cmd)
            output = ''

        if self.json_decode and not skip_json_decode:
            return CmdOutput(json.loads(output))

        if self.yaml_decode:
            return CmdOutput(yaml.safe_load(output))

        if self.singleline:
            return CmdOutput(output.strip())

        return CmdOutput(output.splitlines(keepends=True))


class FileCmd(FileCmdBase):
    """ Implements file-based command execution.

    This is used e.g. with sosreports where the output of a command is saved
    to disk.
    """
    TYPE = "FILE"

    @catch_exceptions(*CLI_COMMON_EXCEPTIONS)
    @reset_command
    @run_post_exec_hooks
    @run_pre_exec_hooks
    def __call__(self, *args, skip_load_contents=False, **kwargs):
        if args:
            self.path = self.path.format(*args)

        if kwargs:
            self.path = self.path.format(**kwargs)

        if not os.path.exists(self.path):
            raise SourceNotFound(self.path)

        if skip_load_contents:
            CmdOutput(None, self.path)

        # NOTE: any post-exec hooks must be aware that their input will be
        # defined by the following.
        if self.json_decode:
            with open(self.path, encoding='utf-8') as fd:
                output = json.load(fd)
        elif self.yaml_decode:
            with open(self.path, encoding='utf-8') as fd:
                output = yaml.safe_.load(fd)
        else:
            output = []
            ln = 0
            with open(self.path, 'rb') as fd:
                for line in fd:
                    ln += 1
                    try:
                        decode_kwargs = {}
                        if self.decode_error_handling:
                            decode_kwargs['errors'] = \
                                                     self.decode_error_handling

                        _out = line.decode(**decode_kwargs)
                        output.append(_out)
                    except UnicodeDecodeError:
                        # maintain line count but store empty line.
                        # we could in the future consider other decode options
                        # as a fallback.
                        output.append('')
                        log.exception("failed to decode line %s "
                                      "(decode_error_handling=%s)", ln,
                                      self.decode_error_handling)

                    if self.singleline:
                        break

            if self.singleline:
                return CmdOutput(output[0].strip(), self.path)

        return CmdOutput(output, self.path)


class BinFileCmd(FileCmd):
    """ This is used when we are executing an actual binary/command against a
    file. """

    @catch_exceptions(*CLI_COMMON_EXCEPTIONS)
    @reset_command
    @run_post_exec_hooks
    @run_pre_exec_hooks
    def __call__(self, *args, **kwargs):
        # NOTE: we check the original path since by this point the 'path'
        # attribute will have been modified to include a binary and other args
        # required to execute it.
        if not os.path.exists(self.get_original_attr_value('path')):
            raise SourceNotFound(self.get_original_attr_value('path'))

        if args:
            self.path = self.path.format(*args)

        if kwargs:
            self.path = self.path.format(**kwargs)

        # If this file is part of a sosreport we want to make sure it is run
        # in the same timezone context as the sosreport host.
        env = {}
        try:
            env['TZ'] = DateFileCmd('sos_commands/date/date',
                                    singleline=True)(format="+%Z").value
        except SourceNotFound:
            pass

        # Now split into a command and run
        output = subprocess.check_output(self.path.split(),
                                         timeout=HotSOSConfig.command_timeout,
                                         stderr=subprocess.STDOUT, env=env)

        output = output.decode('UTF-8')
        return CmdOutput(output.splitlines(keepends=True))


class DateFileCmd(FileCmd):
    """ Implementation of file-based date command. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.register_hook("post-exec", self.format_date)

    def format_date(self, output, **kwargs):
        """ Apply some post-processing to the date output.

        @param output: CmdOutput object
        """
        no_format = kwargs.get('no_format', False)
        fmt = kwargs.get('format')
        if not no_format and fmt is None:
            fmt = '+%s'

        ret = re.match(r"^(\S+ \S*\s*[0-9]+ [0-9:]+)\s*"
                       r"([A-Z]*|[+-]?[0-9]*)?"
                       r"\s*([0-9]+)$",
                       output.value)

        if ret is None:
            log.error("%s has invalid date string '%s'", self.path,
                      output.value)
            return CmdOutput('', self.path)

        tz = ret[2]
        # NOTE: date command doesn't recognise HKT for some reason so we
        # convert to a format that is recognised.
        if tz == 'HKT':
            tz = 'UTC+8'

        # Include tz if available and then convert to utc
        date = f"{ret[1]} {tz} {ret[3]}"
        cmd = ["date", "--utc", f"--date={date}"]
        if fmt:
            cmd.append(fmt)

        output = subprocess.check_output(cmd,
                                         timeout=HotSOSConfig.command_timeout)
        # date sometimes adds multiple whitespaces between fields so collapse
        # them.
        output = re.compile(r"\s+").sub(' ', output.decode('UTF-8'))
        ret = output.splitlines(keepends=True)[0]
        # always singleline so always strip trailing newline
        return CmdOutput(ret.strip(), self.path)



from functools import cached_property

from hotsos.core.log import log
from hotsos.core.host_helpers import (
    APTPackageHelper,
    DPKGVersion,
)
from hotsos.core.ycheck.engine.properties.requires import (
    intercept_exception,
    YRequirementTypeBase,
    PackageCheckItemsBase,
)


class APTCheckItems(PackageCheckItemsBase):
    """
    Implementation for check items on apt package properties.
    """
    @cached_property
    def packaging_helper(self):
        return APTPackageHelper(self.packages_to_check)

    @cached_property
    def installed_versions(self):
        return [self.packaging_helper.get_version(p) for p in self.installed]


class YRequirementTypeAPT(YRequirementTypeBase):
    """
    Apt requires type property. Provides support for defining the apt package
    requires type to perform checks on APT packages.
    """
    override_keys = ['apt']
    override_autoregister = True

    @property
    @intercept_exception
    def _result(self):
        # pylint: disable=duplicate-code
        _result = True
        items = APTCheckItems(self.content)

        # bail on first fail i.e. if any not installed
        if not items.not_installed:
            for pkg, versions in items:
                log.debug("package %s installed=%s", pkg, _result)
                if not versions:
                    continue
                pkg_version = items.packaging_helper.get_version(pkg)
                _result = DPKGVersion.is_version_within_ranges(pkg_version,
                                                               versions)
                # bail at first failure
                if not _result:
                    break
        else:
            log.debug("one or more packages not installed so returning False "
                      "- %s", ', '.join(items.not_installed))
            # bail on first fail i.e. if any not installed
            _result = False

        self.cache.set('package', ', '.join(items.installed))
        self.cache.set('version', ', '.join([
            str(x) for x in items.installed_versions]))
        log.debug('requirement check: apt %s (result=%s)',
                  ', '.join(items.packages_to_check), _result)
        return _result
